# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Utilities for calculating and applying affine orientations"""

import numpy as np
import numpy.linalg as npl

from .deprecated import deprecate_with_version


class OrientationError(Exception):
    pass


def io_orientation(affine, tol=None):
    """Orientation of input axes in terms of output axes for `affine`

    Valid for an affine transformation from ``p`` dimensions to ``q``
    dimensions (``affine.shape == (q + 1, p + 1)``).

    The calculated orientations can be used to transform associated
    arrays to best match the output orientations. If ``p`` > ``q``, then
    some of the output axes should be considered dropped in this
    orientation.

    Parameters
    ----------
    affine : (q+1, p+1) ndarray-like
       Transformation affine from ``p`` inputs to ``q`` outputs.  Usually this
       will be a shape (4,4) matrix, transforming 3 inputs to 3 outputs, but
       the code also handles the more general case
    tol : {None, float}, optional
       threshold below which SVD values of the affine are considered zero. If
       `tol` is None, and ``S`` is an array with singular values for `affine`,
       and ``eps`` is the epsilon value for datatype of ``S``, then `tol` set
       to ``S.max() * max((q, p)) * eps``

    Returns
    -------
    orientations : (p, 2) ndarray
       one row per input axis, where the first value in each row is the closest
       corresponding output axis. The second value in each row is 1 if the
       input axis is in the same direction as the corresponding output axis and
       -1 if it is in the opposite direction.  If a row is [np.nan, np.nan],
       which can happen when p > q, then this row should be considered dropped.
    """
    affine = np.asarray(affine)
    q, p = affine.shape[0] - 1, affine.shape[1] - 1
    # extract the underlying rotation, zoom, shear matrix
    RZS = affine[:q, :p]
    zooms = np.sqrt(np.sum(RZS * RZS, axis=0))
    # Zooms can be zero, in which case all elements in the column are zero, and
    # we can leave them as they are
    zooms[zooms == 0] = 1
    RS = RZS / zooms
    # Transform below is polar decomposition, returning the closest
    # shearless matrix R to RS
    P, S, Qs = npl.svd(RS, full_matrices=False)
    # Threshold the singular values to determine the rank.
    if tol is None:
        tol = S.max() * max(RS.shape) * np.finfo(S.dtype).eps
    keep = S > tol
    R = np.dot(P[:, keep], Qs[keep])
    # the matrix R is such that np.dot(R,R.T) is projection onto the
    # columns of P[:,keep] and np.dot(R.T,R) is projection onto the rows
    # of Qs[keep].  R (== np.dot(R, np.eye(p))) gives rotation of the
    # unit input vectors to output coordinates.  Therefore, the row
    # index of abs max R[:,N], is the output axis changing most as input
    # axis N changes.  In case there are ties, we choose the axes
    # iteratively, removing used axes from consideration as we go
    ornt = np.ones((p, 2), dtype=np.int8) * np.nan
    # Process input axes from strongest to weakest (stable on ties) so a given
    # dimension is labeled consistently regardless of the original order.
    in_axes = np.argsort(np.min(-(R**2), axis=0), kind='stable')
    for in_ax in in_axes:
        col = R[:, in_ax]
        if not np.allclose(col, 0):
            out_ax = np.argmax(np.abs(col))
            ornt[in_ax, 0] = out_ax
            assert col[out_ax] != 0
            if col[out_ax] < 0:
                ornt[in_ax, 1] = -1
            else:
                ornt[in_ax, 1] = 1
            # remove the identified axis from further consideration, by
            # zeroing out the corresponding row in R
            R[out_ax, :] = 0
    return ornt


def ornt_transform(start_ornt, end_ornt):
    """Return the orientation that transforms from `start_ornt` to `end_ornt`.

    Parameters
    ----------
    start_ornt : (n,2) orientation array
        Initial orientation.

    end_ornt : (n,2) orientation array
        Final orientation.

    Returns
    -------
    orientations : (p, 2) ndarray
       The orientation that will transform the `start_ornt` to the `end_ornt`.
    """
    start_ornt = np.asarray(start_ornt)
    end_ornt = np.asarray(end_ornt)
    if start_ornt.shape != end_ornt.shape:
        raise ValueError('The orientations must have the same shape')
    if start_ornt.shape[1] != 2:
        raise ValueError(f'Invalid shape for an orientation: {start_ornt.shape}')
    result = np.empty_like(start_ornt)
    for end_in_idx, (end_out_idx, end_flip) in enumerate(end_ornt):
        for start_in_idx, (start_out_idx, start_flip) in enumerate(start_ornt):
            if end_out_idx == start_out_idx:
                if start_flip == end_flip:
                    flip = 1
                else:
                    flip = -1
                result[start_in_idx, :] = [end_in_idx, flip]
                break
        else:
            raise ValueError(f'Unable to find out axis {end_out_idx} in start_ornt')
    return result


def apply_orientation(arr, ornt):
    """Apply transformations implied by `ornt` to the first
    n axes of the array `arr`

    Parameters
    ----------
    arr : array-like of data with ndim >= n
    ornt : (n,2) orientation array
       orientation transform. ``ornt[N,1]` is flip of axis N of the
       array implied by `shape`, where 1 means no flip and -1 means
       flip.  For example, if ``N==0`` and ``ornt[0,1] == -1``, and
       there's an array ``arr`` of shape `shape`, the flip would
       correspond to the effect of ``np.flipud(arr)``.  ``ornt[:,0]`` is
       the transpose that needs to be done to the implied array, as in
       ``arr.transpose(ornt[:,0])``

    Returns
    -------
    t_arr : ndarray
       data array `arr` transformed according to ornt
    """
    t_arr = np.asarray(arr)
    ornt = np.asarray(ornt)
    n = ornt.shape[0]
    if t_arr.ndim < n:
        raise OrientationError('Data array has fewer dimensions than orientation')
    # no coordinates can be dropped for applying the orientations
    if np.any(np.isnan(ornt[:, 0])):
        raise OrientationError('Cannot drop coordinates when applying orientation to data')
    # apply ornt transformations
    for ax, flip in enumerate(ornt[:, 1]):
        if flip == -1:
            t_arr = np.flip(t_arr, axis=ax)
    full_transpose = np.arange(t_arr.ndim)
    # ornt indicates the transpose that has occurred - we reverse it
    full_transpose[:n] = np.argsort(ornt[:, 0])
    t_arr = t_arr.transpose(full_transpose)
    return t_arr


def inv_ornt_aff(ornt, shape):
    """Affine transform reversing transforms implied in `ornt`

    Imagine you have an array ``arr`` of shape `shape`, and you apply the
    transforms implied by `ornt` (more below), to get ``tarr``.
    ``tarr`` may have a different shape ``shape_prime``.  This routine
    returns the affine that will take a array coordinate for ``tarr``
    and give you the corresponding array coordinate in ``arr``.

    Parameters
    ----------
    ornt : (p, 2) ndarray
       orientation transform. ``ornt[P, 1]` is flip of axis N of the array
       implied by `shape`, where 1 means no flip and -1 means flip.  For
       example, if ``P==0`` and ``ornt[0, 1] == -1``, and there's an array
       ``arr`` of shape `shape`, the flip would correspond to the effect of
       ``np.flipud(arr)``.  ``ornt[:,0]`` gives us the (reverse of the)
       transpose that has been done to ``arr``.  If there are any NaNs in
       `ornt`, we raise an ``OrientationError`` (see notes)
    shape : length p sequence
       shape of array you may transform with `ornt`

    Returns
    -------
    transform_affine : (p + 1, p + 1) ndarray
       An array ``arr`` (shape `shape`) might be transformed according to
       `ornt`, resulting in a transformed array ``tarr``.  `transformed_affine`
       is the transform that takes you from array coordinates in ``tarr`` to
       array coordinates in ``arr``.

    Notes
    -----
    If a row in `ornt` contains NaN, this means that the input row does not
    influence the output space, and is thus effectively dropped from the output
    space.  In that case one ``tarr`` coordinate maps to many ``arr``
    coordinates, we can't invert the transform, and we raise an error
    """
    ornt = np.asarray(ornt)
    if np.any(np.isnan(ornt)):
        raise OrientationError('We cannot invert orientation transform')
    p = ornt.shape[0]
    shape = np.array(shape)[:p]
    # ornt implies a flip, followed by a transpose.   We need the affine
    # that inverts these.  Thus we need the affine that first undoes the
    # effect of the transpose, then undoes the effects of the flip.
    # ornt indicates the transpose that has occurred to get the current
    # ordering, relative to canonical, so we just use that.
    # undo_reorder is a row permutatation matrix
    axis_transpose = [int(v) for v in ornt[:, 0]]
    undo_reorder = np.eye(p + 1)[axis_transpose + [p], :]
    undo_flip = np.diag(list(ornt[:, 1]) + [1.0])
    center_trans = -(shape - 1) / 2.0
    undo_flip[:p, p] = (ornt[:, 1] * center_trans) - center_trans
    return np.dot(undo_flip, undo_reorder)


@deprecate_with_version(
    'flip_axis is deprecated. Please use numpy.flip instead.',
    '3.2',
    '5.0',
)
def flip_axis(arr, axis=0):
    """Flip contents of `axis` in array `arr`

    Equivalent to ``np.flip(arr, axis)``.

    Parameters
    ----------
    arr : array-like
    axis : int, optional
       axis to flip.  Default `axis` == 0

    Returns
    -------
    farr : array
       Array with axis `axis` flipped
    """
    return np.flip(arr, axis)


def ornt2axcodes(ornt, labels=None):
    """Convert orientation `ornt` to labels for axis directions

    Parameters
    ----------
    ornt : (N,2) array-like
        orientation array - see io_orientation docstring
    labels : optional, None or sequence of (2,) sequences
        (2,) sequences are labels for (beginning, end) of output axis.  That
        is, if the first row in `ornt` is ``[1, 1]``, and the second (2,)
        sequence in `labels` is ('back', 'front') then the first returned axis
        code will be ``'front'``.  If the first row in `ornt` had been
        ``[1, -1]`` then the first returned value would have been ``'back'``.
        If None, equivalent to ``(('L','R'),('P','A'),('I','S'))`` - that is -
        RAS axes.

    Returns
    -------
    axcodes : (N,) tuple
        labels for positive end of voxel axes.  Dropped axes get a label of
        None.

    Examples
    --------
    >>> ornt2axcodes([[1, 1],[0,-1],[2,1]], (('L','R'),('B','F'),('D','U')))
    ('F', 'L', 'U')
    """
    if labels is None:
        labels = list(zip('LPI', 'RAS'))
    axcodes = []
    for axno, direction in np.asarray(ornt):
        if np.isnan(axno):
            axcodes.append(None)
            continue
        axint = int(np.round(axno))
        if axint != axno:
            raise ValueError(f'Non integer axis number {axno:f}')
        elif direction == 1:
            axcode = labels[axint][1]
        elif direction == -1:
            axcode = labels[axint][0]
        else:
            raise ValueError('Direction should be -1 or 1')
        axcodes.append(axcode)
    return tuple(axcodes)


def axcodes2ornt(axcodes, labels=None):
    """Convert axis codes `axcodes` to an orientation

    Parameters
    ----------
    axcodes : (N,) tuple
        axis codes - see ornt2axcodes docstring
    labels : optional, None or sequence of (2,) sequences
        (2,) sequences are labels for (beginning, end) of output axis.  That
        is, if the first element in `axcodes` is ``front``, and the second
        (2,) sequence in `labels` is ('back', 'front') then the first
        row of `ornt` will be ``[1, 1]``. If None, equivalent to
        ``(('L','R'),('P','A'),('I','S'))`` - that is - RAS axes.

    Returns
    -------
    ornt : (N,2) array-like
        orientation array - see io_orientation docstring

    Examples
    --------
    >>> axcodes2ornt(('F', 'L', 'U'), (('L','R'),('B','F'),('D','U')))
    array([[ 1.,  1.],
           [ 0., -1.],
           [ 2.,  1.]])
    """
    labels = list(zip('LPI', 'RAS')) if labels is None else labels
    allowed_labels = sum(map(list, labels), [None])
    if len(allowed_labels) != len(set(allowed_labels)):
        raise ValueError(f'Duplicate labels in {allowed_labels}')
    if not set(axcodes).issubset(allowed_labels):
        raise ValueError(f'Not all axis codes {list(axcodes)} in label set {allowed_labels}')
    n_axes = len(axcodes)
    ornt = np.ones((n_axes, 2), dtype=np.int8) * np.nan
    for code_idx, code in enumerate(axcodes):
        for label_idx, codes in enumerate(labels):
            if code is None:
                continue
            if code in codes:
                if code == codes[0]:
                    ornt[code_idx, :] = [label_idx, -1]
                else:
                    ornt[code_idx, :] = [label_idx, 1]
                break
    return ornt


def aff2axcodes(aff, labels=None, tol=None):
    """axis direction codes for affine `aff`

    Parameters
    ----------
    aff : (N,M) array-like
        affine transformation matrix
    labels : optional, None or sequence of (2,) sequences
        Labels for negative and positive ends of output axes of `aff`.  See
        docstring for ``ornt2axcodes`` for more detail
    tol : None or float
        Tolerance for SVD of affine - see ``io_orientation`` for more detail.

    Returns
    -------
    axcodes : (N,) tuple
        labels for positive end of voxel axes.  Dropped axes get a label of
        None.

    Examples
    --------
    >>> aff = [[0,1,0,10],[-1,0,0,20],[0,0,1,30],[0,0,0,1]]
    >>> aff2axcodes(aff, (('L','R'),('B','F'),('D','U')))
    ('B', 'R', 'U')
    """
    ornt = io_orientation(aff, tol)
    return ornt2axcodes(ornt, labels)



"""Test floating point deconstructions and floor methods"""

import sys

import numpy as np
from packaging.version import Version

from ..casting import (
    FloatingError,
    _check_maxexp,
    _check_nmant,
    ceil_exact,
    floor_exact,
    have_binary128,
    longdouble_precision_improved,
    ok_floats,
    on_powerpc,
    sctypes,
    type_info,
)
from ..testing import suppress_warnings

IEEE_floats = [np.float16, np.float32, np.float64]

LD_INFO = type_info(np.longdouble)

FP_OVERFLOW_WARN = Version(np.__version__) < Version('2.0.0.dev0')


def dtt2dict(dtt):
    """Create info dictionary from numpy type"""
    info = np.finfo(dtt)
    return dict(
        min=info.min,
        max=info.max,
        nexp=info.nexp,
        nmant=info.nmant,
        minexp=info.minexp,
        maxexp=info.maxexp,
        width=np.dtype(dtt).itemsize,
    )


def test_type_info():
    # Test routine to get min, max, nmant, nexp
    for dtt in sctypes['int'] + sctypes['uint']:
        info = np.iinfo(dtt)
        infod = type_info(dtt)
        assert infod == dict(
            min=info.min,
            max=info.max,
            nexp=None,
            nmant=None,
            minexp=None,
            maxexp=None,
            width=np.dtype(dtt).itemsize,
        )
        assert infod['min'].dtype.type == dtt
        assert infod['max'].dtype.type == dtt
    for dtt in IEEE_floats + [np.complex64, np.complex64]:
        infod = type_info(dtt)
        assert dtt2dict(dtt) == infod
        assert infod['min'].dtype.type == dtt
        assert infod['max'].dtype.type == dtt
    # What is longdouble?
    ld_dict = dtt2dict(np.longdouble)
    dbl_dict = dtt2dict(np.float64)
    infod = type_info(np.longdouble)
    vals = tuple(ld_dict[k] for k in ('nmant', 'nexp', 'width'))
    # Information for PPC head / tail doubles from:
    # https://developer.apple.com/library/mac/#documentation/Darwin/Reference/Manpages/man3/float.3.html
    if vals in (
        (52, 11, 8),  # longdouble is same as double
        (63, 15, 12),  # intel 80 bit
        (63, 15, 16),  # intel 80 bit
        (112, 15, 16),  # real float128
        (106, 11, 16),  # PPC head, tail doubles, expected values
    ):
        pass
    elif vals == (105, 11, 16):  # bust info for PPC head / tail longdoubles
        # min and max broken, copy from infod
        ld_dict.update({k: infod[k] for k in ('min', 'max')})
    elif vals == (1, 1, 16):  # another bust info for PPC head / tail longdoubles
        ld_dict = dbl_dict.copy()
        ld_dict.update(dict(nmant=106, width=16))
    elif vals == (52, 15, 12):
        width = ld_dict['width']
        ld_dict = dbl_dict.copy()
        ld_dict['width'] = width
    else:
        raise ValueError(f'Unexpected float type {np.longdouble} to test')
    assert ld_dict == infod


def test_nmant():
    for t in IEEE_floats:
        assert type_info(t)['nmant'] == np.finfo(t).nmant
    if (LD_INFO['nmant'], LD_INFO['nexp']) == (63, 15):
        assert type_info(np.longdouble)['nmant'] == 63


def test_check_nmant_nexp():
    # Routine for checking number of sigificand digits and exponent
    for t in IEEE_floats:
        nmant = np.finfo(t).nmant
        maxexp = np.finfo(t).maxexp
        assert _check_nmant(t, nmant)
        assert not _check_nmant(t, nmant - 1)
        assert not _check_nmant(t, nmant + 1)
        with suppress_warnings():  # overflow
            assert _check_maxexp(t, maxexp)
        assert not _check_maxexp(t, maxexp - 1)
        with suppress_warnings():
            assert not _check_maxexp(t, maxexp + 1)
    # Check against type_info
    for t in ok_floats():
        ti = type_info(t)
        if ti['nmant'] not in (105, 106):  # This check does not work for PPC double pair
            assert _check_nmant(t, ti['nmant'])
        # Test fails for longdouble after blacklisting of macOS powl as of numpy
        # 1.12 - see https://github.com/numpy/numpy/issues/8307
        if t != np.longdouble or sys.platform != 'darwin':
            assert _check_maxexp(t, ti['maxexp'])


def test_int_longdouble_np_regression():
    # Test longdouble conversion from int works as expected
    # Previous versions of numpy would fail, and we used a custom int_to_float()
    # function. This test remains to ensure we don't need to bring it back.
    nmant = type_info(np.float64)['nmant']
    # test we recover precision just above nmant
    i = 2 ** (nmant + 1) - 1
    assert int(np.longdouble(i)) == i
    assert int(np.longdouble(-i)) == -i
    # If longdouble can cope with 2**64, test
    if nmant >= 63:
        # Check conversion to int; the line below causes an error subtracting
        # ints / uint64 values, at least for Python 3.3 and numpy dev 1.8
        big_int = np.uint64(2**64 - 1)
        assert int(np.longdouble(big_int)) == big_int


def test_int_np_regression():
    # Test int works as expected for integers.
    # We previously used a custom as_int() for integers because of a
    # numpy 1.4.1 bug such that int(np.uint32(2**32-1) == -1
    for t in sctypes['int'] + sctypes['uint']:
        info = np.iinfo(t)
        mn, mx = np.array([info.min, info.max], dtype=t)
        assert (mn, mx) == (int(mn), int(mx))


def test_floor_exact_16():
    # A normal integer can generate an inf in float16
    assert floor_exact(2**31, np.float16) == np.inf
    assert floor_exact(-(2**31), np.float16) == -np.inf


def test_floor_exact_64():
    # float64
    for e in range(53, 63):
        start = np.float64(2**e)
        across = start + np.arange(2048, dtype=np.float64)
        gaps = set(np.diff(across)).difference([0])
        assert len(gaps) == 1
        gap = gaps.pop()
        assert gap == int(gap)
        test_val = 2 ** (e + 1) - 1
        assert floor_exact(test_val, np.float64) == 2 ** (e + 1) - int(gap)


def test_floor_exact(max_digits):
    max_digits(4950)  # max longdouble is ~10**4932

    to_test = IEEE_floats + [float]
    try:
        type_info(np.longdouble)['nmant']
    except FloatingError:
        # Significand bit count not reliable, don't test long double
        pass
    else:
        to_test.append(np.longdouble)
    # When numbers go above int64 - I believe, numpy comparisons break down,
    # so we have to cast to int before comparison
    int_flex = lambda x, t: int(floor_exact(x, t))
    int_ceex = lambda x, t: int(ceil_exact(x, t))
    for t in to_test:
        # A number bigger than the range returns the max
        info = type_info(t)
        assert floor_exact(10**4933, t) == np.inf
        assert ceil_exact(10**4933, t) == np.inf
        # A number more negative returns -inf
        assert floor_exact(-(10**4933), t) == -np.inf
        assert ceil_exact(-(10**4933), t) == -np.inf
        # Check around end of integer precision
        nmant = info['nmant']
        for i in range(nmant + 1):
            iv = 2**i
            # up to 2**nmant should be exactly representable
            for func in (int_flex, int_ceex):
                assert func(iv, t) == iv
                assert func(-iv, t) == -iv
                assert func(iv - 1, t) == iv - 1
                assert func(-iv + 1, t) == -iv + 1
        if t is np.longdouble and (on_powerpc() or longdouble_precision_improved()):
            # The nmant value for longdouble on PPC appears to be conservative,
            # so that the tests for behavior above the nmant range fail.
            # windows longdouble can change from float64 to Intel80 in some
            # situations, in which case nmant will not be correct
            continue
        # Confirm to ourselves that 2**(nmant+1) can't be exactly represented
        iv = 2 ** (nmant + 1)
        assert int_flex(iv + 1, t) == iv
        assert int_ceex(iv + 1, t) == iv + 2
        # negatives
        assert int_flex(-iv - 1, t) == -iv - 2
        assert int_ceex(-iv - 1, t) == -iv
        # The gap in representable numbers is 2 above 2**(nmant+1), 4 above
        # 2**(nmant+2), and so on.
        for i in range(5):
            iv = 2 ** (nmant + 1 + i)
            gap = 2 ** (i + 1)
            assert int(t(iv) + t(gap)) == iv + gap
            for j in range(1, gap):
                assert int_flex(iv + j, t) == iv
                assert int_flex(iv + gap + j, t) == iv + gap
                assert int_ceex(iv + j, t) == iv + gap
                assert int_ceex(iv + gap + j, t) == iv + 2 * gap
            # negatives
            for j in range(1, gap):
                assert int_flex(-iv - j, t) == -iv - gap
                assert int_flex(-iv - gap - j, t) == -iv - 2 * gap
                assert int_ceex(-iv - j, t) == -iv
                assert int_ceex(-iv - gap - j, t) == -iv - gap


def test_usable_binary128():
    # Check for usable binary128
    yes = have_binary128()
    with np.errstate(over='ignore'):
        exp_test = np.longdouble(2) ** 16383
    assert yes == (
        exp_test.dtype.itemsize == 16
        and np.isfinite(exp_test)
        and _check_nmant(np.longdouble, 112)
    )



"""Testing Siemens CSA header reader"""

import struct
import sys

from ..structreader import Unpacker


def test_unpacker():
    s = b'1234\x00\x01'
    (le_int,) = struct.unpack('<h', b'\x00\x01')
    (be_int,) = struct.unpack('>h', b'\x00\x01')
    if sys.byteorder == 'little':
        native_int = le_int
        swapped_int = be_int
        swapped_code = '>'
    else:
        native_int = be_int
        swapped_int = le_int
        swapped_code = '<'
    up_str = Unpacker(s, endian='<')
    assert up_str.read(4) == b'1234'
    up_str.ptr = 0
    assert up_str.unpack('4s') == (b'1234',)
    assert up_str.unpack('h') == (le_int,)
    up_str = Unpacker(s, endian='>')
    assert up_str.unpack('4s') == (b'1234',)
    assert up_str.unpack('h') == (be_int,)
    # now test conflict of endian
    up_str = Unpacker(s, ptr=4, endian='>')
    assert up_str.unpack('<h') == (le_int,)
    up_str = Unpacker(s, ptr=4, endian=swapped_code)
    assert up_str.unpack('h') == (swapped_int,)
    up_str.ptr = 4
    assert up_str.unpack('<h') == (le_int,)
    up_str.ptr = 4
    assert up_str.unpack('>h') == (be_int,)
    up_str.ptr = 4
    assert up_str.unpack('@h') == (native_int,)
    # test -1 for read
    up_str.ptr = 2
    assert up_str.read() == b'34\x00\x01'
    # past end
    assert up_str.read() == b''
    # with n_bytes
    up_str.ptr = 2
    assert up_str.read(2) == b'34'
    assert up_str.read(2) == b'\x00\x01'



"""Adapter module for working with pydicom < 1.0 and >= 1.0

In what follows, "dicom is available" means we can import either a) ``dicom``
(pydicom < 1.0) or or b) ``pydicom`` (pydicom >= 1.0).

Regardless of whether dicom is available this module should be importable
without error, and always defines:

* have_dicom : True if we can import pydicom or dicom;
* pydicom : pydicom module or dicom module or None if not importable;
* read_file : ``read_file`` function if pydicom or dicom module is importable
  else None;
* tag_for_keyword : ``tag_for_keyword`` function if pydicom or dicom module
  is importable else None;

A test decorator is available in nibabel.nicom.tests:

* dicom_test : test decorator that skips test if dicom not available.

A deprecated copy is available here for backward compatibility.
"""

from __future__ import annotations

import warnings
from collections.abc import Callable

from .deprecated import deprecate_with_version
from .optpkg import optional_package

warnings.warn(
    "We will remove the 'pydicom_compat' module from nibabel 7.0. "
    "Please consult pydicom's documentation for any future needs.",
    DeprecationWarning,
    stacklevel=2,
)

pydicom, have_dicom, _ = optional_package('pydicom')

read_file: Callable | None = None
tag_for_keyword: Callable | None = None
Sequence: type | None = None

if have_dicom:
    # Values not imported by default
    import pydicom.values  # type: ignore[import-not-found]
    from pydicom.dicomio import dcmread as read_file  # noqa:F401
    from pydicom.sequence import Sequence  # noqa:F401

    tag_for_keyword = pydicom.datadict.tag_for_keyword


@deprecate_with_version(
    'dicom_test has been moved to nibabel.nicom.tests', since='3.1', until='5.0'
)
def dicom_test(func):
    # Import locally to avoid circular dependency
    from .nicom.tests import dicom_test

    return dicom_test(func)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""DICOM reader

.. currentmodule:: nibabel.nicom

.. autosummary::
   :toctree: ../generated

   csareader
   dicomreaders
   dicomwrappers
   dwiparams
   structreader
"""

import warnings

warnings.warn(
    'The DICOM readers are highly experimental, unstable,'
    ' and only work for Siemens time-series at the moment\n'
    'Please use with caution.  We would be grateful for your '
    'help in improving them',
    UserWarning,
    stacklevel=2,
)



"""Testing reading DICOM files"""

from os.path import join as pjoin

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from nibabel.optpkg import optional_package

from .. import dicomreaders as didr
from .test_dicomwrappers import DATA, EXPECTED_AFFINE, EXPECTED_PARAMS, IO_DATA_PATH

pydicom, _, setup_module = optional_package('pydicom')


def test_read_dwi():
    img = didr.mosaic_to_nii(DATA)
    arr = img.get_fdata()
    assert arr.shape == (128, 128, 48)
    assert_array_almost_equal(img.affine, EXPECTED_AFFINE)


def test_read_dwis():
    data, aff, bs, gs = didr.read_mosaic_dwi_dir(IO_DATA_PATH, 'siemens_dwi_*.dcm.gz')
    assert data.ndim == 4
    assert_array_almost_equal(aff, EXPECTED_AFFINE)
    assert_array_almost_equal(bs, (0, EXPECTED_PARAMS[0]))
    assert_array_almost_equal(gs, (np.zeros((3,)), EXPECTED_PARAMS[1]))
    with pytest.raises(OSError):
        didr.read_mosaic_dwi_dir('improbable')


def test_passing_kwds():
    # Check that we correctly pass keywords to dicom
    dwi_glob = 'siemens_dwi_*.dcm.gz'
    csa_glob = 'csa*.bin'
    for func in (didr.read_mosaic_dwi_dir, didr.read_mosaic_dir):
        data, aff, bs, gs = func(IO_DATA_PATH, dwi_glob)
        # This should not raise an error
        data2, aff2, bs2, gs2 = func(IO_DATA_PATH, dwi_glob, dicom_kwargs=dict(force=True))
        assert_array_equal(data, data2)
        # This should raise an error in pydicom.filereader.dcmread
        with pytest.raises(TypeError):
            func(IO_DATA_PATH, dwi_glob, dicom_kwargs=dict(not_a_parameter=True))
        # These are invalid dicoms, so will raise an error unless force=True
        with pytest.raises(pydicom.filereader.InvalidDicomError):
            func(IO_DATA_PATH, csa_glob)
        # But here, we catch the error because the dicoms are in the wrong
        # format
        with pytest.raises(didr.DicomReadError):
            func(IO_DATA_PATH, csa_glob, dicom_kwargs=dict(force=True))


def test_slices_to_series():
    dicom_files = (pjoin(IO_DATA_PATH, f'{i}.dcm') for i in range(2))
    wrappers = [didr.wrapper_from_file(f) for f in dicom_files]
    series = didr.slices_to_series(wrappers)
    assert len(series) == 1
    assert len(series[0]) == 2



"""Testing imageclasses module"""

from os.path import dirname
from os.path import join as pjoin

import numpy as np

import nibabel as nib
from nibabel.analyze import AnalyzeImage
from nibabel.imageclasses import spatial_axes_first
from nibabel.nifti1 import Nifti1Image
from nibabel.nifti2 import Nifti2Image
from nibabel.optpkg import optional_package

have_h5py = optional_package('h5py')[1]

DATA_DIR = pjoin(dirname(__file__), 'data')

MINC_3DS = ('minc1_1_scale.mnc',)
MINC_4DS = ('minc1_4d.mnc',)
if have_h5py:
    MINC_3DS = MINC_3DS + ('minc2_1_scale.mnc',)
    MINC_4DS = MINC_4DS + ('minc2_4d.mnc',)


def test_spatial_axes_first():
    # Function tests is spatial axes are first three axes in image
    # Always True for Nifti and friends
    affine = np.eye(4)
    for shape in ((2, 3), (4, 3, 2), (5, 4, 1, 2), (2, 3, 5, 2, 1)):
        for img_class in (AnalyzeImage, Nifti1Image, Nifti2Image):
            data = np.zeros(shape)
            img = img_class(data, affine)
            assert spatial_axes_first(img)
    # True for MINC images < 4D
    for fname in MINC_3DS:
        img = nib.load(pjoin(DATA_DIR, fname))
        assert len(img.shape) == 3
        assert spatial_axes_first(img)
    # False for MINC images < 4D
    for fname in MINC_4DS:
        img = nib.load(pjoin(DATA_DIR, fname))
        assert len(img.shape) == 4
        assert not spatial_axes_first(img)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import base64
import os.path as op
import sys
import warnings
import zlib
from io import StringIO
from xml.parsers.expat import ExpatError

import numpy as np

from ..nifti1 import data_type_codes, intent_codes, xform_codes
from ..xmlutils import XmlParser
from .gifti import (
    GiftiCoordSystem,
    GiftiDataArray,
    GiftiImage,
    GiftiLabel,
    GiftiLabelTable,
    GiftiMetaData,
)
from .util import array_index_order_codes, gifti_encoding_codes, gifti_endian_codes


class GiftiParseError(ExpatError):
    """Gifti-specific parsing error"""


def read_data_block(darray, fname, data, mmap):
    """Parses data from a <Data> element, or loads from an external file.

    Parameters
    ----------
    darray : GiftiDataArray
         GiftiDataArray object representing the parent <DataArray> of this
         <Data> element

    fname : str or None
         Name of GIFTI file being loaded, or None if in-memory

    data : str or None
         Data to parse, or None if data is in an external file

    mmap : {True, False, 'c', 'r', 'r+'}
        Controls the use of numpy memory mapping for reading data.  Only has
        an effect when loading GIFTI images with data stored in external files
        (``DataArray`` elements with an ``Encoding`` equal to
        ``ExternalFileBinary``).  If ``False``, do not try numpy ``memmap``
        for data array.  If one of ``{'c', 'r', 'r+'}``, try numpy ``memmap``
        with ``mode=mmap``.  A `mmap` value of ``True`` gives the same
        behavior as ``mmap='c'``.  If the file cannot be memory-mapped, ignore
        `mmap` value and read array from file.

    Returns
    -------
    ``numpy.ndarray`` or ``numpy.memmap`` containing the parsed data
    """
    if mmap not in (True, False, 'c', 'r', 'r+'):
        raise ValueError("mmap value should be one of True, False, 'c', 'r', 'r+'")
    if mmap is True:
        mmap = 'c'
    enclabel = gifti_encoding_codes.label[darray.encoding]

    if enclabel not in ('ASCII', 'B64BIN', 'B64GZ', 'External'):
        raise GiftiParseError(f'Unknown encoding {darray.encoding}')

    # Encode the endianness in the dtype
    byteorder = gifti_endian_codes.byteorder[darray.endian]
    dtype = data_type_codes.dtype[darray.datatype].newbyteorder(byteorder)

    shape = tuple(darray.dims)
    order = array_index_order_codes.npcode[darray.ind_ord]

    # GIFTI_ENCODING_ASCII
    if enclabel == 'ASCII':
        return np.loadtxt(StringIO(data), dtype=dtype, ndmin=1).reshape(shape, order=order)

    # We assume that the external data file is raw uncompressed binary, with
    # the data type/endianness/ordering specified by the other DataArray
    # attributes
    if enclabel == 'External':
        if fname is None:
            raise GiftiParseError(
                'ExternalFileBinary is not supported when loading from in-memory XML'
            )
        ext_fname = op.join(op.dirname(fname), darray.ext_fname)
        if not op.exists(ext_fname):
            raise GiftiParseError('Cannot locate external file ' + ext_fname)
        # We either create a memmap, or load into memory
        newarr = None
        if mmap:
            try:
                return np.memmap(
                    ext_fname,
                    dtype=dtype,
                    mode=mmap,
                    offset=darray.ext_offset,
                    shape=shape,
                    order=order,
                )
            # If the memmap fails, we ignore the error and load the data into
            # memory below
            except (AttributeError, TypeError, ValueError):
                pass
        # mmap=False or np.memmap failed
        if newarr is None:
            return np.fromfile(
                ext_fname,
                dtype=dtype,
                count=np.prod(darray.dims),
                offset=darray.ext_offset,
            ).reshape(shape, order=order)

    # Numpy arrays created from bytes objects are read-only.
    # Neither b64decode nor decompress will return bytearrays, and there
    # are not equivalents to fobj.readinto to allow us to pass them, so
    # there is not a simple way to avoid making copies.
    # If this becomes a problem, we should write a decoding interface with
    # a tunable chunk size.
    dec = base64.b64decode(data.encode('ascii'))
    if enclabel == 'B64BIN':
        buff = bytearray(dec)
    else:
        # GIFTI_ENCODING_B64GZ
        buff = bytearray(zlib.decompress(dec))
    del dec
    return np.frombuffer(buff, dtype=dtype).reshape(shape, order=order)


def _str2int(in_str):
    # Convert string to integer, where empty string gives 0
    return int(in_str) if in_str else 0


class GiftiImageParser(XmlParser):
    def __init__(self, encoding=None, buffer_size=35000000, verbose=0, mmap=True):
        super().__init__(encoding=encoding, buffer_size=buffer_size, verbose=verbose)
        # output
        self.img = None

        # Queried when loading data from <Data> elements - see read_data_block
        self.mmap = mmap

        # finite state machine stack
        self.fsm_state = []

        # temporary constructs
        self.nvpair = None
        self.da = None
        self.coordsys = None
        self.lata = None
        self.label = None

        self.meta_global = None
        self.meta_da = None
        self.count_da = True

        # where to write CDATA:
        self.write_to = None

        # Collecting char buffer fragments
        self._char_blocks = None

    def StartElementHandler(self, name, attrs):
        self.flush_chardata()
        if self.verbose > 0:
            print('Start element:\n\t', repr(name), attrs)

        if name == 'GIFTI':
            # create gifti image
            self.img = GiftiImage()
            if 'Version' in attrs:
                self.img.version = attrs['Version']
            if 'NumberOfDataArrays' in attrs:
                self.expected_numDA = int(attrs['NumberOfDataArrays'])
            self.fsm_state.append('GIFTI')

        elif name == 'MetaData':
            self.fsm_state.append('MetaData')
            # if this metadata tag is first, create self.img.meta
            if len(self.fsm_state) == 2:
                self.meta_global = GiftiMetaData()
            else:
                # otherwise, create darray.meta
                self.meta_da = GiftiMetaData()

        elif name == 'MD':
            self.nvpair = ['', '']
            self.fsm_state.append('MD')

        elif name == 'Name':
            if self.nvpair is None:
                raise GiftiParseError
            self.write_to = 'Name'

        elif name == 'Value':
            if self.nvpair is None:
                raise GiftiParseError
            self.write_to = 'Value'

        elif name == 'LabelTable':
            self.lata = GiftiLabelTable()
            self.fsm_state.append('LabelTable')

        elif name == 'Label':
            self.label = GiftiLabel()
            if 'Index' in attrs:
                self.label.key = int(attrs['Index'])
            if 'Key' in attrs:
                self.label.key = int(attrs['Key'])
            if 'Red' in attrs:
                self.label.red = float(attrs['Red'])
            if 'Green' in attrs:
                self.label.green = float(attrs['Green'])
            if 'Blue' in attrs:
                self.label.blue = float(attrs['Blue'])
            if 'Alpha' in attrs:
                self.label.alpha = float(attrs['Alpha'])
            self.write_to = 'Label'

        elif name == 'DataArray':
            self.da = GiftiDataArray()
            if 'Intent' in attrs:
                self.da.intent = intent_codes.code[attrs['Intent']]
            if 'DataType' in attrs:
                self.da.datatype = data_type_codes.code[attrs['DataType']]
            if 'ArrayIndexingOrder' in attrs:
                self.da.ind_ord = array_index_order_codes.code[attrs['ArrayIndexingOrder']]
            num_dim = int(attrs.get('Dimensionality', 0))
            for i in range(num_dim):
                di = f'Dim{i}'
                if di in attrs:
                    self.da.dims.append(int(attrs[di]))
            # dimensionality has to correspond to the number of DimX given
            # TODO (bcipolli): don't assert; raise parse warning, and recover.
            assert len(self.da.dims) == num_dim
            if 'Encoding' in attrs:
                self.da.encoding = gifti_encoding_codes.code[attrs['Encoding']]
            if 'Endian' in attrs:
                self.da.endian = gifti_endian_codes.code[attrs['Endian']]
            if 'ExternalFileName' in attrs:
                self.da.ext_fname = attrs['ExternalFileName']
            if 'ExternalFileOffset' in attrs:
                self.da.ext_offset = _str2int(attrs['ExternalFileOffset'])
            self.img.darrays.append(self.da)
            self.fsm_state.append('DataArray')

        elif name == 'CoordinateSystemTransformMatrix':
            self.coordsys = GiftiCoordSystem()
            self.img.darrays[-1].coordsys = self.coordsys
            self.fsm_state.append('CoordinateSystemTransformMatrix')

        elif name == 'DataSpace':
            if self.coordsys is None:
                raise GiftiParseError
            self.write_to = 'DataSpace'

        elif name == 'TransformedSpace':
            if self.coordsys is None:
                raise GiftiParseError
            self.write_to = 'TransformedSpace'

        elif name == 'MatrixData':
            if self.coordsys is None:
                raise GiftiParseError
            self.write_to = 'MatrixData'

        elif name == 'Data':
            self.write_to = 'Data'

    def EndElementHandler(self, name):
        self.flush_chardata()
        if self.verbose > 0:
            print('End element:\n\t', repr(name))

        if name == 'GIFTI':
            if hasattr(self, 'expected_numDA') and self.expected_numDA != self.img.numDA:
                warnings.warn(
                    'Actual # of data arrays does not match # expected: '
                    f'{self.expected_numDA} != {self.img.numDA}.'
                )
            # remove last element of the list
            self.fsm_state.pop()
            # assert len(self.fsm_state) == 0

        elif name == 'MetaData':
            self.fsm_state.pop()
            if len(self.fsm_state) == 1:
                # only Gifti there, so this was a closing global
                # metadata tag
                self.img.meta = self.meta_global
                self.meta_global = None
            else:
                self.img.darrays[-1].meta = self.meta_da
                self.meta_da = None

        elif name == 'MD':
            self.fsm_state.pop()
            key, val = self.nvpair
            if self.meta_global is not None and self.meta_da is None:
                self.meta_global[key] = val
            elif self.meta_da is not None and self.meta_global is None:
                self.meta_da[key] = val
            # remove reference
            self.nvpair = None

        elif name == 'LabelTable':
            self.fsm_state.pop()
            # add labeltable
            self.img.labeltable = self.lata
            self.lata = None

        elif name == 'DataArray':
            self.fsm_state.pop()

        elif name == 'CoordinateSystemTransformMatrix':
            self.fsm_state.pop()
            self.coordsys = None

        elif name in ('DataSpace', 'TransformedSpace', 'MatrixData', 'Name', 'Value', 'Data'):
            self.write_to = None

        elif name == 'Label':
            self.lata.labels.append(self.label)
            self.label = None
            self.write_to = None

    def CharacterDataHandler(self, data):
        """Collect character data chunks pending collation

        The parser breaks the data up into chunks of size depending on the
        buffer_size of the parser.  A large bit of character data, with
        standard parser buffer_size (such as 8K) can easily span many calls to
        this function.  We thus collect the chunks and process them when we
        hit start or end tags.
        """
        if self._char_blocks is None:
            self._char_blocks = []
        self._char_blocks.append(data)

    def flush_chardata(self):
        """Collate and process collected character data"""
        # Nothing to do for empty elements, except for Data elements which
        # are within a DataArray with an external file
        if self.write_to != 'Data' and self._char_blocks is None:
            return
        # Just join the strings to get the data.  Maybe there are some memory
        # optimizations we could do by passing the list of strings to the
        # read_data_block function.
        if self._char_blocks is not None:
            data = ''.join(self._char_blocks)
        else:
            data = None
        # Reset the char collector
        self._char_blocks = None

        # Process data
        if self.write_to == 'Name':
            data = data.strip()
            self.nvpair[0] = data

        elif self.write_to == 'Value':
            data = data.strip()
            self.nvpair[1] = data

        elif self.write_to == 'DataSpace':
            data = data.strip()
            self.coordsys.dataspace = xform_codes.code[data]

        elif self.write_to == 'TransformedSpace':
            data = data.strip()
            self.coordsys.xformspace = xform_codes.code[data]

        elif self.write_to == 'MatrixData':
            # conversion to numpy array
            c = StringIO(data)
            self.coordsys.xform = np.loadtxt(c)
            c.close()

        elif self.write_to == 'Data':
            self.da.data = read_data_block(self.da, self.fname, data, self.mmap)
            # update the endianness according to the
            # current machine setting
            self.endian = gifti_endian_codes.code[sys.byteorder]

        elif self.write_to == 'Label':
            self.label.label = data.strip()

    @property
    def pending_data(self):
        """True if there is character data pending for processing"""
        return self._char_blocks is not None



"""Define static nibabel metadata for nibabel

The long description parameter is used in the nibabel top-level docstring,
and in building the docs.
We exec this file in several places, so it cannot import nibabel or use
relative imports.
"""

# Note: this long_description is the canonical place to edit this text.
# It also appears in README.rst, but it should get there by running
# ``tools/refresh_readme.py`` which pulls in this version.
# We also include this text in the docs by ``..include::`` in
# ``docs/source/index.rst``.
long_description = """
Read and write access to common neuroimaging file formats, including:
ANALYZE_ (plain, SPM99, SPM2 and later), GIFTI_, NIfTI1_, NIfTI2_, `CIFTI-2`_,
MINC1_, MINC2_, `AFNI BRIK/HEAD`_, ECAT_ and Philips PAR/REC.
In addition, NiBabel also supports FreeSurfer_'s MGH_, geometry, annotation and
morphometry files, and provides some limited support for DICOM_.

NiBabel's API gives full or selective access to header information (metadata),
and image data is made available via NumPy arrays. For more information, see
NiBabel's `documentation site`_ and `API reference`_.

.. _API reference: https://nipy.org/nibabel/api.html
.. _AFNI BRIK/HEAD: https://afni.nimh.nih.gov/pub/dist/src/README.attributes
.. _ANALYZE: http://www.grahamwideman.com/gw/brain/analyze/formatdoc.htm
.. _CIFTI-2: https://www.nitrc.org/projects/cifti/
.. _DICOM: http://medical.nema.org/
.. _documentation site: http://nipy.org/nibabel
.. _ECAT: http://xmedcon.sourceforge.net/Docs/Ecat
.. _Freesurfer: https://surfer.nmr.mgh.harvard.edu
.. _GIFTI: https://www.nitrc.org/projects/gifti
.. _MGH: https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
.. _MINC1:
    https://en.wikibooks.org/wiki/MINC/Reference/MINC1_File_Format_Reference
.. _MINC2:
    https://en.wikibooks.org/wiki/MINC/Reference/MINC2.0_File_Format_Reference
.. _NIfTI1: http://nifti.nimh.nih.gov/nifti-1/
.. _NIfTI2: http://nifti.nimh.nih.gov/nifti-2/

Installation
============

To install NiBabel's `current release`_ with ``pip``, run::

   pip install nibabel

To install the latest development version, run::

   pip install git+https://github.com/nipy/nibabel

When working on NiBabel itself, it may be useful to install in "editable" mode::

   git clone https://github.com/nipy/nibabel.git
   pip install -e ./nibabel

For more information on previous releases, see the `release archive`_ or
`development changelog`_.

.. _current release: https://pypi.python.org/pypi/NiBabel
.. _release archive: https://github.com/nipy/NiBabel/releases
.. _development changelog: https://nipy.org/nibabel/changelog.html

Testing
=======

During development, we recommend using tox_ to run nibabel tests::

    git clone https://github.com/nipy/nibabel.git
    cd nibabel
    tox

To test an installed version of nibabel, install the test dependencies
and run pytest_::

    pip install nibabel[test]
    pytest --pyargs nibabel

For more information, consult the `developer guidelines`_.

.. _tox: https://tox.wiki
.. _pytest: https://docs.pytest.org
.. _developer guidelines: https://nipy.org/nibabel/devel/devguide.html

Mailing List
============

Please send any questions or suggestions to the `neuroimaging mailing list
<https://mail.python.org/mailman/listinfo/neuroimaging>`_.

License
=======

NiBabel is licensed under the terms of the `MIT license
<https://github.com/nipy/nibabel/blob/master/COPYING#nibabel>`__.
Some code included with NiBabel is licensed under the `BSD license`_.
For more information, please see the COPYING_ file.

.. _BSD license: https://opensource.org/licenses/BSD-3-Clause
.. _COPYING: https://github.com/nipy/nibabel/blob/master/COPYING

Citation
========

NiBabel releases have a Zenodo_ `Digital Object Identifier`_ (DOI) badge at
the top of the release notes. Click on the badge for more information.

.. _Digital Object Identifier: https://en.wikipedia.org/wiki/Digital_object_identifier
.. _zenodo: https://zenodo.org
"""



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Read / write access to SPM2 version of analyze image format"""

import numpy as np

from . import spm99analyze as spm99  # module import

image_dimension_dtd = spm99.image_dimension_dtd.copy()
image_dimension_dtd[image_dimension_dtd.index(('funused2', 'f4'))] = ('scl_inter', 'f4')

# Full header numpy dtype combined across sub-fields
header_dtype = np.dtype(spm99.header_key_dtd + image_dimension_dtd + spm99.data_history_dtd)


class Spm2AnalyzeHeader(spm99.Spm99AnalyzeHeader):
    """Class for SPM2 variant of basic Analyze header

    SPM2 variant adds the following to basic Analyze format:

    * voxel origin;
    * slope scaling of data;
    * reading - but not writing - intercept of data.
    """

    # Copies of module level definitions
    template_dtype = header_dtype

    def get_slope_inter(self):
        """Get data scaling (slope) and intercept from header data

        Uses the algorithm from SPM2 spm_vol_ana.m by John Ashburner

        Parameters
        ----------
        self : header
           Mapping with fields:
           * scl_slope - slope
           * scl_inter - possible intercept (SPM2 use - shared by nifti)
           * glmax - the (recorded) maximum value in the data (unscaled)
           * glmin - recorded minimum unscaled value
           * cal_max - the calibrated (scaled) maximum value in the dataset
           * cal_min - ditto minimum value

        Returns
        -------
        scl_slope : None or float
            slope.  None if there is no valid scaling from these fields
        scl_inter : None or float
            intercept.  Also None if there is no valid slope, intercept

        Examples
        --------
        >>> fields = {'scl_slope': 1, 'scl_inter': 0, 'glmax': 0, 'glmin': 0,
        ...           'cal_max': 0, 'cal_min': 0}
        >>> hdr = Spm2AnalyzeHeader()
        >>> for key, value in fields.items():
        ...     hdr[key] = value
        >>> hdr.get_slope_inter()
        (1.0, 0.0)
        >>> hdr['scl_inter'] = 0.5
        >>> hdr.get_slope_inter()
        (1.0, 0.5)
        >>> hdr['scl_inter'] = np.nan
        >>> hdr.get_slope_inter()
        (1.0, 0.0)

        If 'scl_slope' is 0, nan or inf, cannot use 'scl_slope'.
        Without valid information in the gl / cal fields, we cannot get
        scaling, and return None

        >>> hdr['scl_slope'] = 0
        >>> hdr.get_slope_inter()
        (None, None)
        >>> hdr['scl_slope'] = np.nan
        >>> hdr.get_slope_inter()
        (None, None)

        Valid information in the gl AND cal fields are needed

        >>> hdr['cal_max'] = 0.8
        >>> hdr['cal_min'] = 0.2
        >>> hdr.get_slope_inter()
        (None, None)
        >>> hdr['glmax'] = 110
        >>> hdr['glmin'] = 10
        >>> np.allclose(hdr.get_slope_inter(), [0.6/100, 0.2-0.6/100*10])
        True
        """
        # get scaling factor from 'scl_slope' (funused1)
        slope = float(self['scl_slope'])
        if np.isfinite(slope) and slope:
            # try to get offset from scl_inter
            inter = float(self['scl_inter'])
            if not np.isfinite(inter):
                inter = 0.0
            return slope, inter
        # no non-zero and finite scaling, try gl/cal fields
        unscaled_range = self['glmax'] - self['glmin']
        scaled_range = self['cal_max'] - self['cal_min']
        if unscaled_range and scaled_range:
            slope = float(scaled_range) / unscaled_range
            inter = self['cal_min'] - slope * self['glmin']
            return slope, inter
        return None, None

    @classmethod
    def may_contain_header(klass, binaryblock):
        if len(binaryblock) < klass.sizeof_hdr:
            return False

        hdr_struct = np.ndarray(
            shape=(), dtype=header_dtype, buffer=binaryblock[: klass.sizeof_hdr]
        )
        bs_hdr_struct = hdr_struct.byteswap()
        return binaryblock[344:348] not in (b'ni1\x00', b'n+1\x00') and 348 in (
            hdr_struct['sizeof_hdr'],
            bs_hdr_struct['sizeof_hdr'],
        )


class Spm2AnalyzeImage(spm99.Spm99AnalyzeImage):
    """Class for SPM2 variant of basic Analyze image"""

    header_class = Spm2AnalyzeHeader
    header: Spm2AnalyzeHeader


load = Spm2AnalyzeImage.from_filename
save = Spm2AnalyzeImage.instance_to_filename



"""Testing loadsave module"""

import pathlib
import shutil
from os.path import dirname
from os.path import join as pjoin
from tempfile import TemporaryDirectory

import numpy as np

from .. import (
    Nifti1Image,
    Nifti1Pair,
    Nifti2Image,
    Nifti2Pair,
    Spm2AnalyzeImage,
    Spm99AnalyzeImage,
)
from .._compression import HAVE_ZSTD
from ..filebasedimages import ImageFileError
from ..loadsave import _signature_matches_extension, load, read_img_data
from ..openers import Opener
from ..optpkg import optional_package
from ..testing import deprecated_to, expires
from ..tmpdirs import InTemporaryDirectory

_, have_scipy, _ = optional_package('scipy')

import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

data_path = pjoin(dirname(__file__), 'data')


@expires('5.0.0')
def test_read_img_data():
    fnames_test = [
        'example4d.nii.gz',
        'example_nifti2.nii.gz',
        'minc1_1_scale.mnc',
        'minc1_4d.mnc',
        'test.mgz',
        'tiny.mnc',
    ]
    fnames_test += [pathlib.Path(p) for p in fnames_test]
    for fname in fnames_test:
        fpath = pjoin(data_path, fname)
        if isinstance(fname, pathlib.Path):
            fpath = pathlib.Path(fpath)
        img = load(fpath)
        data = img.get_fdata()
        with deprecated_to('5.0.0'):
            data2 = read_img_data(img)
        assert_array_equal(data, data2)
        # These examples have null scaling - assert prefer=unscaled is the same
        dao = img.dataobj
        if hasattr(dao, 'slope') and hasattr(img.header, 'raw_data_from_fileobj'):
            assert (dao.slope, dao.inter) == (1, 0)
            with deprecated_to('5.0.0'):
                assert_array_equal(read_img_data(img, prefer='unscaled'), data)
        # Assert all caps filename works as well
        with TemporaryDirectory() as tmpdir:
            up_fpath = pjoin(tmpdir, str(fname).upper())
            if isinstance(fname, pathlib.Path):
                up_fpath = pathlib.Path(up_fpath)
            shutil.copyfile(fpath, up_fpath)
            img = load(up_fpath)
            assert_array_equal(img.dataobj, data)
            del img


def test_file_not_found():
    with pytest.raises(FileNotFoundError):
        load('does_not_exist.nii.gz')


def test_load_empty_image():
    with InTemporaryDirectory():
        open('empty.nii', 'w').close()
        with pytest.raises(ImageFileError) as err:
            load('empty.nii')
    assert str(err.value).startswith('Empty file: ')


@pytest.mark.parametrize('extension', ['.gz', '.bz2', '.zst'])
def test_load_bad_compressed_extension(tmp_path, extension):
    if extension == '.zst' and not HAVE_ZSTD:
        pytest.skip()
    file_path = tmp_path / f'img.nii{extension}'
    file_path.write_bytes(b'bad')
    with pytest.raises(ImageFileError, match=r'.*is not a .* file'):
        load(file_path)


@pytest.mark.parametrize('extension', ['.gz', '.bz2', '.zst'])
def test_load_good_extension_with_bad_data(tmp_path, extension):
    if extension == '.zst' and not HAVE_ZSTD:
        pytest.skip()
    file_path = tmp_path / f'img.nii{extension}'
    with Opener(file_path, 'wb') as fobj:
        fobj.write(b'bad')
    with pytest.raises(ImageFileError, match=r'Cannot work out file type of .*'):
        load(file_path)


def test_signature_matches_extension(tmp_path):
    gz_signature = b'\x1f\x8b'
    good_file = tmp_path / 'good.gz'
    good_file.write_bytes(gz_signature)
    bad_file = tmp_path / 'bad.gz'
    bad_file.write_bytes(b'bad')
    matches, msg = _signature_matches_extension(tmp_path / 'uncompressed.nii')
    assert matches
    assert msg == ''
    matches, msg = _signature_matches_extension(tmp_path / 'missing.gz')
    assert not matches
    assert msg.startswith('Could not read')
    matches, msg = _signature_matches_extension(bad_file)
    assert not matches
    assert 'is not a' in msg
    matches, msg = _signature_matches_extension(good_file)
    assert matches
    assert msg == ''
    matches, msg = _signature_matches_extension(tmp_path / 'missing.nii')
    assert matches
    assert msg == ''


@expires('5.0.0')
def test_read_img_data_nifti():
    shape = (2, 3, 4)
    data = np.random.normal(size=shape)
    out_dtype = np.dtype(np.int16)
    classes = (Nifti1Pair, Nifti1Image, Nifti2Pair, Nifti2Image)
    if have_scipy:
        classes += (Spm99AnalyzeImage, Spm2AnalyzeImage)
    with InTemporaryDirectory():
        for i, img_class in enumerate(classes):
            img = img_class(data, np.eye(4))
            img.set_data_dtype(out_dtype)
            # No filemap => error
            with deprecated_to('5.0.0'), pytest.raises(ImageFileError):
                read_img_data(img)
            # Make a filemap
            froot = f'an_image_{i}'
            img.file_map = img.filespec_to_file_map(froot)
            # Trying to read from this filemap will generate an error because
            # we are going to read from files that do not exist
            with deprecated_to('5.0.0'), pytest.raises(OSError):
                read_img_data(img)
            img.to_file_map()
            # Load - now the scaling and offset correctly applied
            img_fname = img.file_map['image'].filename
            img_back = load(img_fname)
            data_back = img_back.get_fdata()
            with deprecated_to('5.0.0'):
                assert_array_equal(data_back, read_img_data(img_back))
            # This is the same as if we loaded the image and header separately
            hdr_fname = img.file_map['header'].filename if 'header' in img.file_map else img_fname
            with open(hdr_fname, 'rb') as fobj:
                hdr_back = img_back.header_class.from_fileobj(fobj)
            with open(img_fname, 'rb') as fobj:
                scaled_back = hdr_back.data_from_fileobj(fobj)
            assert_array_equal(data_back, scaled_back)
            # Unscaled is the same as returned from raw_data_from_fileobj
            with open(img_fname, 'rb') as fobj:
                unscaled_back = hdr_back.raw_data_from_fileobj(fobj)
            with deprecated_to('5.0.0'):
                assert_array_equal(unscaled_back, read_img_data(img_back, prefer='unscaled'))
            # If we futz with the scaling in the header, the result changes
            with deprecated_to('5.0.0'):
                assert_array_equal(data_back, read_img_data(img_back))
            has_inter = hdr_back.has_data_intercept
            old_slope = hdr_back['scl_slope']
            old_inter = hdr_back['scl_inter'] if has_inter else 0
            est_unscaled = (data_back - old_inter) / old_slope
            with deprecated_to('5.0.0'):
                actual_unscaled = read_img_data(img_back, prefer='unscaled')
            assert_almost_equal(est_unscaled, actual_unscaled)
            img_back.header['scl_slope'] = 2.1
            if has_inter:
                new_inter = 3.14
                img_back.header['scl_inter'] = 3.14
            else:
                new_inter = 0
            # scaled scaling comes from new parameters in header
            with deprecated_to('5.0.0'):
                assert np.allclose(actual_unscaled * 2.1 + new_inter, read_img_data(img_back))
            # Unscaled array didn't change
            with deprecated_to('5.0.0'):
                assert_array_equal(actual_unscaled, read_img_data(img_back, prefer='unscaled'))
            # Check the offset too
            img.header.set_data_offset(1024)
            # Delete arrays still pointing to file, so Windows can reuse
            del actual_unscaled, unscaled_back
            img.to_file_map()
            # Write an integer of zeros after
            with open(img_fname, 'ab') as fobj:
                fobj.write(b'\x00\x00')
            img_back = load(img_fname)
            data_back = img_back.get_fdata()
            with deprecated_to('5.0.0'):
                assert_array_equal(data_back, read_img_data(img_back))
            img_back.header.set_data_offset(1026)
            # Check we pick up new offset
            exp_offset = np.zeros((data.size,), data.dtype) + old_inter
            exp_offset[:-1] = np.ravel(data_back, order='F')[1:]
            exp_offset = np.reshape(exp_offset, shape, order='F')
            with deprecated_to('5.0.0'):
                assert_array_equal(exp_offset, read_img_data(img_back))
            # Delete stuff that might hold onto file references
            del img, img_back, data_back



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import bz2
import gzip
from io import BytesIO
from os.path import join as pjoin

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from .. import Nifti1Image, load, minc1
from .._compression import HAVE_ZSTD, zstd
from ..externals.netcdf import netcdf_file
from ..minc1 import Minc1File, Minc1Image, MincHeader
from ..testing import assert_data_similar, data_path
from ..tmpdirs import InTemporaryDirectory
from . import test_spatialimages as tsi
from .test_fileslice import slicer_samples

EG_FNAME = pjoin(data_path, 'tiny.mnc')

# Example images in format expected for ``test_image_api``, adding ``zooms``
# item.
EXAMPLE_IMAGES = [
    dict(
        fname=pjoin(data_path, 'tiny.mnc'),
        shape=(10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 2.0, -20],
                [0, 2.0, 0, -20],
                [2.0, 0, 0, -10],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(2.0, 2.0, 2.0),
        # These values from SPM2
        data_summary=dict(min=0.20784314, max=0.74901961, mean=0.60602819),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc1_1_scale.mnc'),
        shape=(10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 2.0, -20],
                [0, 2.0, 0, -20],
                [2.0, 0, 0, -10],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(2.0, 2.0, 2.0),
        # These values from mincstats
        data_summary=dict(min=0.2082842439, max=0.2094327615, mean=0.2091292083),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc1_4d.mnc'),
        shape=(2, 10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 2.0, -20],
                [0, 2.0, 0, -20],
                [2.0, 0, 0, -10],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(1.0, 2.0, 2.0, 2.0),
        # These values from mincstats
        data_summary=dict(min=0.2078431373, max=1.498039216, mean=0.9090422837),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc1-no-att.mnc'),
        shape=(10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 1.0, 0],
                [0, 1.0, 0, 0],
                [1.0, 0, 0, 0],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(1.0, 1.0, 1.0),
        # These values from SPM2/mincstats
        data_summary=dict(min=0.20784314, max=0.74901961, mean=0.6061103),
        is_proxy=True,
    ),
]


class _TestMincFile:
    module = minc1
    file_class = Minc1File
    fname = EG_FNAME
    opener = netcdf_file
    test_files = EXAMPLE_IMAGES

    def test_mincfile(self):
        for tp in self.test_files:
            mnc_obj = self.opener(tp['fname'], 'r')
            mnc = self.file_class(mnc_obj)
            assert mnc.get_data_dtype().type == tp['dtype']
            assert mnc.get_data_shape() == tp['shape']
            assert mnc.get_zooms() == tp['zooms']
            assert_array_equal(mnc.get_affine(), tp['affine'])
            data = mnc.get_scaled_data()
            assert data.shape == tp['shape']
            # Can't close mmapped NetCDF with live mmap arrays
            del mnc, data

    def test_mincfile_slicing(self):
        # Test slicing and scaling of mincfile data
        for tp in self.test_files:
            mnc_obj = self.opener(tp['fname'], 'r')
            mnc = self.file_class(mnc_obj)
            data = mnc.get_scaled_data()
            for slicedef in (
                (slice(None),),
                (1,),
                (slice(None), 1),
                (1, slice(None)),
                (slice(None), 1, 1),
                (1, slice(None), 1),
                (1, 1, slice(None)),
            ):
                sliced_data = mnc.get_scaled_data(slicedef)
                assert_array_equal(sliced_data, data[slicedef])
            # Can't close mmapped NetCDF with live mmap arrays
            del mnc, data

    def test_load(self):
        # Check highest level load of minc works
        for tp in self.test_files:
            img = load(tp['fname'])
            data = img.get_fdata()
            assert data.shape == tp['shape']
            # min, max, mean values from read in SPM2 / minctools
            assert_data_similar(data, tp)
            # check if mnc can be converted to nifti
            ni_img = Nifti1Image.from_image(img)
            assert_array_equal(ni_img.affine, tp['affine'])
            assert_array_equal(ni_img.get_fdata(), data)

    def test_array_proxy_slicing(self):
        # Test slicing of array proxy
        for tp in self.test_files:
            img = load(tp['fname'])
            arr = img.get_fdata()
            prox = img.dataobj
            assert prox.is_proxy
            for sliceobj in slicer_samples(img.shape):
                assert_array_equal(arr[sliceobj], prox[sliceobj])


class TestMinc1File(_TestMincFile):
    def test_compressed(self):
        # we can read minc compressed
        # Not so for MINC2; hence this small sub-class
        for tp in self.test_files:
            content = open(tp['fname'], 'rb').read()
            openers_exts = [(gzip.open, '.gz'), (bz2.BZ2File, '.bz2')]
            if HAVE_ZSTD:  # add .zst to test if installed
                openers_exts += [(zstd.ZstdFile, '.zst')]
            with InTemporaryDirectory():
                for opener, ext in openers_exts:
                    fname = 'test.mnc' + ext
                    fobj = opener(fname, 'wb')
                    fobj.write(content)
                    fobj.close()
                    img = self.module.load(fname)
                    data = img.get_fdata()
                    assert_data_similar(data, tp)
                    del img


# Test the Minc header
def test_header_data_io():
    bio = BytesIO()
    hdr = MincHeader()
    arr = np.arange(24).reshape((2, 3, 4))
    with pytest.raises(NotImplementedError):
        hdr.data_to_fileobj(arr, bio)
    with pytest.raises(NotImplementedError):
        hdr.data_from_fileobj(bio)


class TestMinc1Image(tsi.TestSpatialImage):
    image_class = Minc1Image
    eg_images = (pjoin(data_path, 'tiny.mnc'),)
    module = minc1

    def test_data_to_from_fileobj(self):
        # Check data_from_fileobj of header raises an error
        for fpath in self.eg_images:
            img = self.module.load(fpath)
            bio = BytesIO()
            arr = np.arange(24).reshape((2, 3, 4))
            with pytest.raises(NotImplementedError):
                img.header.data_to_fileobj(arr, bio)
            with pytest.raises(NotImplementedError):
                img.header.data_from_fileobj(bio)



import argparse
import os
import sys

import nibabel as nb


def lossless_slice(img, slicers):
    if not nb.imageclasses.spatial_axes_first(img):
        raise ValueError('Cannot slice an image that is not known to have spatial axes first')

    scaling = hasattr(img.header, 'set_slope_inter')

    data = img.dataobj._get_unscaled(slicers) if scaling else img.dataobj[slicers]
    roi_img = img.__class__(data, affine=img.slicer.slice_affine(slicers), header=img.header)

    if scaling:
        roi_img.header.set_slope_inter(img.dataobj.slope, img.dataobj.inter)
    return roi_img


def parse_slice(crop, allow_step=True):
    if crop is None:
        return slice(None)
    start, stop, *extra = (int(val) if val else None for val in crop.split(':'))
    if len(extra) > 1:
        raise ValueError(f'Cannot parse specification: {crop}')
    if not allow_step and extra and extra[0] not in (1, None):
        raise ValueError(f'Step entry not permitted: {crop}')

    step = extra[0] if extra else None
    if step not in (1, -1, None):
        raise ValueError(f'Downsampling is not supported: {crop}')

    return slice(start, stop, step)


def sanitize(args):
    # Argparse likes to treat "-1:..." as a flag
    return [f' {arg}' if arg[0] == '-' and ':' in arg else arg for arg in args]


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description='Crop images to a region of interest',
        epilog='If a start or stop value is omitted, the start or end of the axis is assumed.',
    )
    parser.add_argument('--version', action='version', version=nb.__version__)
    parser.add_argument(
        '-i', metavar='I1:I2[:-1]', help='Start/stop [flip] along first axis (0-indexed)'
    )
    parser.add_argument(
        '-j', metavar='J1:J2[:-1]', help='Start/stop [flip] along second axis (0-indexed)'
    )
    parser.add_argument(
        '-k', metavar='K1:K2[:-1]', help='Start/stop [flip] along third axis (0-indexed)'
    )
    parser.add_argument('-t', metavar='T1:T2', help='Start/stop along fourth axis (0-indexed)')
    parser.add_argument('in_file', help='Image file to crop')
    parser.add_argument('out_file', help='Output file name')

    opts = parser.parse_args(args=sanitize(args))

    try:
        islice = parse_slice(opts.i)
        jslice = parse_slice(opts.j)
        kslice = parse_slice(opts.k)
        tslice = parse_slice(opts.t, allow_step=False)
    except ValueError as err:
        print(f'Could not parse input arguments. Reason follows.\n{err}')
        return 1

    kwargs = {}
    if os.path.realpath(opts.in_file) == os.path.realpath(opts.out_file):
        kwargs['mmap'] = False
    img = nb.load(opts.in_file, **kwargs)

    slicers = (islice, jslice, kslice, tslice)[: img.ndim]
    expected_shape = nb.fileslice.predict_shape(slicers, img.shape)
    if any(dim == 0 for dim in expected_shape):
        print(f'Cannot take zero-length slices. Predicted shape {expected_shape}.')
        return 1

    try:
        sliced_img = lossless_slice(img, slicers)
    except Exception:
        print('Could not slice image. Full traceback follows.')
        raise
    nb.save(sliced_img, opts.out_file)
    return 0



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Read MINC1 format images"""

from __future__ import annotations

from numbers import Integral

import numpy as np

from .externals.netcdf import netcdf_file
from .fileslice import canonical_slicers
from .spatialimages import SpatialHeader, SpatialImage

_dt_dict = {
    ('b', 'unsigned'): np.uint8,
    ('b', 'signed__'): np.int8,
    ('c', 'unsigned'): 'S1',
    ('h', 'unsigned'): np.uint16,
    ('h', 'signed__'): np.int16,
    ('i', 'unsigned'): np.uint32,
    ('i', 'signed__'): np.int32,
}

# See
# https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#MINC_specific_convenience_functions
_default_dir_cos = {'xspace': [1, 0, 0], 'yspace': [0, 1, 0], 'zspace': [0, 0, 1]}


class MincError(Exception):
    """Error when reading MINC files"""


class Minc1File:
    """Class to wrap MINC1 format opened netcdf object

    Although it has some of the same methods as a ``Header``, we use
    this only when reading a MINC file, to pull out useful header
    information, and for the method of reading the data out
    """

    def __init__(self, mincfile):
        self._mincfile = mincfile
        self._image = mincfile.variables['image']
        self._dim_names = self._image.dimensions
        # The code below will error with vector_dimensions.  See:
        # https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#An_Introduction_to_NetCDF
        # https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#Image_dimensions
        self._dims = [self._mincfile.variables[s] for s in self._dim_names]
        # We don't currently support irregular spacing
        # https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#MINC_specific_convenience_functions
        for dim in self._dims:
            if dim.spacing != b'regular__':
                raise ValueError('Irregular spacing not supported')
        self._spatial_dims = [name for name in self._dim_names if name.endswith('space')]
        # the MINC standard appears to allow the following variables to
        # be undefined.
        # https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#Image_conversion_variables
        # It wasn't immediately obvious what the defaults were.
        self._image_max = self._mincfile.variables['image-max']
        self._image_min = self._mincfile.variables['image-min']

    def _get_dimensions(self, var):
        # Dimensions for a particular variable
        # Differs for MINC1 and MINC2 - see:
        # https://en.wikibooks.org/wiki/MINC/Reference/MINC2.0_File_Format_Reference#Associating_HDF5_dataspaces_with_MINC_dimensions
        return var.dimensions

    def get_data_dtype(self):
        typecode = self._image.typecode()
        if typecode == 'f':
            dtt = np.dtype(np.float32)
        elif typecode == 'd':
            dtt = np.dtype(np.float64)
        else:
            signtype = self._image.signtype.decode('latin-1')
            dtt = _dt_dict[(typecode, signtype)]
        return np.dtype(dtt).newbyteorder('>')

    def get_data_shape(self):
        return self._image.data.shape

    def get_zooms(self):
        """Get real-world sizes of voxels"""
        # zooms must be positive; but steps in MINC can be negative
        return tuple(abs(float(dim.step)) if hasattr(dim, 'step') else 1.0 for dim in self._dims)

    def get_affine(self):
        nspatial = len(self._spatial_dims)
        rot_mat = np.eye(nspatial)
        steps = np.zeros((nspatial,))
        starts = np.zeros((nspatial,))
        dim_names = list(self._dim_names)  # for indexing in loop
        for i, name in enumerate(self._spatial_dims):
            dim = self._dims[dim_names.index(name)]
            rot_mat[:, i] = (
                dim.direction_cosines
                if hasattr(dim, 'direction_cosines')
                else _default_dir_cos[name]
            )
            steps[i] = dim.step if hasattr(dim, 'step') else 1.0
            starts[i] = dim.start if hasattr(dim, 'start') else 0.0
        origin = np.dot(rot_mat, starts)
        aff = np.eye(nspatial + 1)
        aff[:nspatial, :nspatial] = rot_mat * steps
        aff[:nspatial, nspatial] = origin
        return aff

    def _get_valid_range(self):
        """Return valid range for image data

        The valid range can come from the image 'valid_range' or
        image 'valid_min' and 'valid_max', or, failing that, from the
        data type range
        """
        ddt = self.get_data_dtype()
        info = np.iinfo(ddt.type)
        try:
            valid_range = self._image.valid_range
        except AttributeError:
            try:
                valid_range = [self._image.valid_min, self._image.valid_max]
            except AttributeError:
                valid_range = [info.min, info.max]
        if valid_range[0] < info.min or valid_range[1] > info.max:
            raise ValueError('Valid range outside input data type range')
        return np.asarray(valid_range, dtype=np.float64)

    def _get_scalar(self, var):
        """Get scalar value from NetCDF scalar"""
        return var.getValue()

    def _get_array(self, var):
        """Get array from NetCDF array"""
        return var.data

    def _normalize(self, data, sliceobj=()):
        """Apply scaling to image data `data` already sliced with `sliceobj`

        https://en.wikibooks.org/wiki/MINC/Reference/MINC1-programmers-guide#Pixel_values_and_real_values

        MINC normalization uses "image-min" and "image-max" variables to
        map the data from the valid range of the image to the range
        specified by "image-min" and "image-max".

        The "image-max" and "image-min" are variables that describe the
        "max" and "min" of image over some dimensions of "image".

        The usual case is that "image" has dimensions ["zspace", "yspace",
        "xspace"] and "image-max" has dimensions ["zspace"], but there can be
        up to two dimensions for over which scaling is specified.

        Parameters
        ----------
        data : ndarray
            data after applying `sliceobj` slicing to full image
        sliceobj : tuple, optional
            slice definition. If not specified, assume no slicing has been
            applied to `data`
        """
        ddt = self.get_data_dtype()
        if np.issubdtype(ddt.type, np.floating):
            return data
        image_max = self._image_max
        image_min = self._image_min
        mx_dims = self._get_dimensions(image_max)
        mn_dims = self._get_dimensions(image_min)
        if mx_dims != mn_dims:
            raise MincError('"image-max" and "image-min" do not have the same dimensions')
        nscales = len(mx_dims)
        if nscales > 2:
            raise MincError('More than two scaling dimensions')
        if mx_dims != self._dim_names[:nscales]:
            raise MincError('image-max and image dimensions do not match')
        dmin, dmax = self._get_valid_range()
        out_data = np.clip(data, dmin, dmax)
        if nscales == 0:  # scalar values
            imax = self._get_scalar(image_max)
            imin = self._get_scalar(image_min)
        else:  # 1D or 2D array of scaling values
            # We need to get the correct values from image-max and image-min to
            # do the scaling.
            shape = self.get_data_shape()
            sliceobj = canonical_slicers(sliceobj, shape)
            # Indices into sliceobj referring to image axes
            ax_inds = [i for i, obj in enumerate(sliceobj) if obj is not None]
            assert len(ax_inds) == len(shape)
            # Slice imax, imin using same slicer as for data
            nscales_ax = ax_inds[nscales]
            i_slicer = sliceobj[:nscales_ax]
            # Fill slicer to broadcast against sliced data; add length 1 axis
            # for each axis except int axes (which are dropped by slicing)
            broad_part = tuple(
                None for s in sliceobj[ax_inds[nscales] :] if not isinstance(s, Integral)
            )
            i_slicer += broad_part
            imax = self._get_array(image_max)[i_slicer]
            imin = self._get_array(image_min)[i_slicer]
        slope = (imax - imin) / (dmax - dmin)
        inter = imin - dmin * slope
        out_data *= slope
        out_data += inter
        return out_data

    def get_scaled_data(self, sliceobj=()):
        """Return scaled data for slice definition `sliceobj`

        Parameters
        ----------
        sliceobj : tuple, optional
            slice definition. If not specified, return whole array

        Returns
        -------
        scaled_arr : array
            array from minc file with scaling applied
        """
        if sliceobj == ():
            raw_data = self._image.data
        else:
            raw_data = self._image.data[sliceobj]
        dtype = self.get_data_dtype()
        data = np.asarray(raw_data).view(dtype)
        return self._normalize(data, sliceobj)


class MincImageArrayProxy:
    """MINC implementation of array proxy protocol

    The array proxy allows us to freeze the passed fileobj and
    header such that it returns the expected data array.
    """

    def __init__(self, minc_file):
        self.minc_file = minc_file
        self._shape = minc_file.get_data_shape()

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def is_proxy(self):
        return True

    def __array__(self, dtype=None):
        """Read data from file and apply scaling, casting to ``dtype``

        If ``dtype`` is unspecified, the dtype is automatically determined.

        Parameters
        ----------
        dtype : numpy dtype specifier, optional
            A numpy dtype specifier specifying the type of the returned array.

        Returns
        -------
        array
            Scaled image data with type `dtype`.
        """
        arr = self.minc_file.get_scaled_data(sliceobj=())
        if dtype is not None:
            arr = arr.astype(dtype, copy=False)
        return arr

    def __getitem__(self, sliceobj):
        """Read slice `sliceobj` of data from file"""
        return self.minc_file.get_scaled_data(sliceobj)


class MincHeader(SpatialHeader):
    """Class to contain header for MINC formats"""

    # We don't use the data layout - this just in case we do later
    data_layout = 'C'

    def data_to_fileobj(self, data, fileobj, rescale=True):
        """See Header class for an implementation we can't use"""
        raise NotImplementedError

    def data_from_fileobj(self, fileobj):
        """See Header class for an implementation we can't use"""
        raise NotImplementedError


class Minc1Header(MincHeader):
    @classmethod
    def may_contain_header(klass, binaryblock):
        return binaryblock[:4] == b'CDF\x01'


class Minc1Image(SpatialImage):
    """Class for MINC1 format images

    The MINC1 image class uses the default header type, rather than a specific
    MINC header type - and reads the relevant information from the MINC file on
    load.
    """

    header_class: type[MincHeader] = Minc1Header
    header: MincHeader
    _meta_sniff_len: int = 4
    valid_exts: tuple[str, ...] = ('.mnc',)
    files_types: tuple[tuple[str, str], ...] = (('image', '.mnc'),)
    _compressed_suffixes: tuple[str, ...] = ('.gz', '.bz2', '.zst')

    makeable = True
    rw = False

    ImageArrayProxy = MincImageArrayProxy

    @classmethod
    def from_file_map(klass, file_map, *, mmap=True, keep_file_open=None):
        # Note that mmap and keep_file_open are included for proper
        with file_map['image'].get_prepare_fileobj() as fobj:
            minc_file = Minc1File(netcdf_file(fobj))
            affine = minc_file.get_affine()
            if affine.shape != (4, 4):
                raise MincError('Image does not have 3 spatial dimensions')
            data_dtype = minc_file.get_data_dtype()
            shape = minc_file.get_data_shape()
            zooms = minc_file.get_zooms()
            header = klass.header_class(data_dtype, shape, zooms)
            data = klass.ImageArrayProxy(minc_file)
        return klass(data, affine, header, extra=None, file_map=file_map)


load = Minc1Image.from_filename



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for mghformat reading writing"""

import io
import os
import pathlib

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_array_equal

from ... import imageglobals
from ...fileholders import FileHolder
from ...openers import ImageOpener
from ...spatialimages import HeaderDataError
from ...testing import data_path
from ...tests import test_spatialimages as tsi
from ...tests import test_wrapstruct as tws
from ...tmpdirs import InTemporaryDirectory
from ...volumeutils import sys_is_le
from ...wrapstruct import WrapStructError
from .. import load, save
from ..mghformat import MGHError, MGHHeader, MGHImage

MGZ_FNAME = os.path.join(data_path, 'test.mgz')

# sample voxel to ras matrix (mri_info --vox2ras)
v2r = np.array(
    [
        [1, 2, 3, -13],
        [2, 3, 1, -11.5],
        [3, 1, 2, -11.5],
        [0, 0, 0, 1],
    ],
    dtype=np.float32,
)
# sample voxel to ras - tkr matrix (mri_info --vox2ras-tkr)
v2rtkr = np.array(
    [
        [-1.0, 0.0, 0.0, 1.5],
        [0.0, 0.0, 1.0, -2.5],
        [0.0, -1.0, 0.0, 2.0],
        [0.0, 0.0, 0.0, 1.0],
    ],
    dtype=np.float32,
)

BIG_CODES = ('>', 'big', 'BIG', 'b', 'be', 'B', 'BE')
LITTLE_CODES = ('<', 'little', 'l', 'le', 'L', 'LE')

if sys_is_le:
    BIG_CODES += ('swapped', 's', 'S', '!')
    LITTLE_CODES += ('native', 'n', 'N', '=', '|', 'i', 'I')
else:
    BIG_CODES += ('native', 'n', 'N', '=', '|', 'i', 'I')
    LITTLE_CODES += ('swapped', 's', 'S', '!')


def test_read_mgh():
    # test.mgz was generated by the following command
    # mri_volsynth --dim 3 4 5 2 --vol test.mgz
    # --cdircos 1 2 3 --rdircos 2 3 1 --sdircos 3 1 2
    # mri_volsynth is a FreeSurfer command
    mgz = load(MGZ_FNAME)

    # header
    h = mgz.header
    assert h['version'] == 1
    assert h['type'] == 3
    assert h['dof'] == 0
    assert h['goodRASFlag'] == 1
    assert_array_equal(h['dims'], [3, 4, 5, 2])
    assert_almost_equal(h['tr'], 2.0)
    assert_almost_equal(h['flip_angle'], 0.0)
    assert_almost_equal(h['te'], 0.0)
    assert_almost_equal(h['ti'], 0.0)
    assert_array_almost_equal(h.get_zooms(), [1, 1, 1, 2])
    assert_array_almost_equal(h.get_vox2ras(), v2r)
    assert_array_almost_equal(h.get_vox2ras_tkr(), v2rtkr)

    # data. will be different for your own mri_volsynth invocation
    v = mgz.get_fdata()
    assert_almost_equal(v[1, 2, 3, 0], -0.3047, 4)
    assert_almost_equal(v[1, 2, 3, 1], 0.0018, 4)


def test_write_mgh():
    # write our data to a tmp file
    v = np.arange(120)
    v = v.reshape((5, 4, 3, 2)).astype(np.float32)
    # form a MGHImage object using data and vox2ras matrix
    img = MGHImage(v, v2r)
    with InTemporaryDirectory():
        save(img, 'tmpsave.mgz')
        # read from the tmp file and see if it checks out
        mgz = load('tmpsave.mgz')
        h = mgz.header
        dat = mgz.get_fdata()
        # Delete loaded image to allow file deletion by windows
        del mgz
    # header
    assert h['version'] == 1
    assert h['type'] == 3
    assert h['dof'] == 0
    assert h['goodRASFlag'] == 1
    assert np.array_equal(h['dims'], [5, 4, 3, 2])
    assert_almost_equal(h['tr'], 0.0)
    assert_almost_equal(h['flip_angle'], 0.0)
    assert_almost_equal(h['te'], 0.0)
    assert_almost_equal(h['ti'], 0.0)
    assert_almost_equal(h['fov'], 0.0)
    assert_array_almost_equal(h.get_vox2ras(), v2r)
    # data
    assert_almost_equal(dat, v, 7)


def test_write_noaffine_mgh():
    # now just save the image without the vox2ras transform
    # and see if it uses the default values to save
    v = np.ones((7, 13, 3, 22), np.uint8)
    # form a MGHImage object using data
    # and the default affine matrix (Note the "None")
    img = MGHImage(v, None)
    with InTemporaryDirectory():
        save(img, 'tmpsave.mgz')
        # read from the tmp file and see if it checks out
        mgz = load('tmpsave.mgz')
        h = mgz.header
        # Delete loaded image to allow file deletion by windows
        del mgz
    # header
    assert h['version'] == 1
    assert h['type'] == 0  # uint8 for mgh
    assert h['dof'] == 0
    assert h['goodRASFlag'] == 1
    assert np.array_equal(h['dims'], [7, 13, 3, 22])
    assert_almost_equal(h['tr'], 0.0)
    assert_almost_equal(h['flip_angle'], 0.0)
    assert_almost_equal(h['te'], 0.0)
    assert_almost_equal(h['ti'], 0.0)
    assert_almost_equal(h['fov'], 0.0)
    # important part -- whether default affine info is stored
    assert_array_almost_equal(h['Mdc'], [[-1, 0, 0], [0, 0, 1], [0, -1, 0]])
    assert_array_almost_equal(h['Pxyz_c'], [0, 0, 0])


def test_set_zooms():
    mgz = load(MGZ_FNAME)
    h = mgz.header
    assert_array_almost_equal(h.get_zooms(), [1, 1, 1, 2])
    h.set_zooms([1, 1, 1, 3])
    assert_array_almost_equal(h.get_zooms(), [1, 1, 1, 3])
    for zooms in (
        (-1, 1, 1, 1),
        (1, -1, 1, 1),
        (1, 1, -1, 1),
        (1, 1, 1, -1),
        (1, 1, 1, 1, 5),
    ):
        with pytest.raises(HeaderDataError):
            h.set_zooms(zooms)
    # smoke test for tr=0
    h.set_zooms((1, 1, 1, 0))


def bad_dtype_mgh():
    """This function raises an MGHError exception because
    float64 is not a valid MGH datatype.
    """
    # try to write an unsigned short and make sure it
    # raises MGHError
    v = np.ones((7, 13, 3, 22), np.float64)
    # form a MGHImage object using data
    # and the default affine matrix (Note the "None")
    MGHImage(v, None)


def test_bad_dtype_mgh():
    # Now test the above function
    with pytest.raises(MGHError):
        bad_dtype_mgh()


def test_filename_exts():
    # Test acceptable filename extensions
    v = np.ones((7, 13, 3, 22), np.uint8)
    # form a MGHImage object using data
    # and the default affine matrix (Note the "None")
    img = MGHImage(v, None)
    # Check if these extensions allow round trip
    for ext in ('.mgh', '.mgz'):
        with InTemporaryDirectory():
            fname = 'tmpname' + ext
            save(img, fname)
            # read from the tmp file and see if it checks out
            img_back = load(fname)
            assert_array_equal(img_back.get_fdata(), v)
            del img_back


def _mgh_rt(img, fobj):
    file_map = {'image': FileHolder(fileobj=fobj)}
    img.to_file_map(file_map)
    return MGHImage.from_file_map(file_map)


def test_header_updating():
    # Don't update the header information if the affine doesn't change.
    # Luckily the test.mgz dataset had a bad set of cosine vectors, so these
    # will be changed if the affine gets updated
    mgz = load(MGZ_FNAME)
    hdr = mgz.header
    # Test against mri_info output
    exp_aff = np.loadtxt(
        io.BytesIO(
            b"""
    1.0000   2.0000   3.0000   -13.0000
    2.0000   3.0000   1.0000   -11.5000
    3.0000   1.0000   2.0000   -11.5000
    0.0000   0.0000   0.0000     1.0000"""
        )
    )
    assert_almost_equal(mgz.affine, exp_aff, 6)
    assert_almost_equal(hdr.get_affine(), exp_aff, 6)
    # Test that initial wonky header elements have not changed
    assert np.all(hdr['delta'] == 1)
    assert_almost_equal(hdr['Mdc'].T, exp_aff[:3, :3])
    # Save, reload, same thing
    img_fobj = io.BytesIO()
    mgz2 = _mgh_rt(mgz, img_fobj)
    hdr2 = mgz2.header
    assert_almost_equal(hdr2.get_affine(), exp_aff, 6)
    assert_array_equal(hdr2['delta'], 1)
    # Change affine, change underlying header info
    exp_aff_d = exp_aff.copy()
    exp_aff_d[0, -1] = -14
    # This will (probably) become part of the official API
    mgz2._affine[:] = exp_aff_d
    mgz2.update_header()
    assert_almost_equal(hdr2.get_affine(), exp_aff_d, 6)
    RZS = exp_aff_d[:3, :3]
    assert_almost_equal(hdr2['delta'], np.sqrt(np.sum(RZS**2, axis=0)))
    assert_almost_equal(hdr2['Mdc'].T, RZS / hdr2['delta'])


def test_cosine_order():
    # Test we are interpreting the cosine order right
    data = np.arange(60, dtype=np.int32).reshape((3, 4, 5))
    aff = np.diag([2.0, 3, 4, 1])
    aff[0] = [2, 1, 0, 10]
    img = MGHImage(data, aff)
    assert_almost_equal(img.affine, aff, 6)
    img_fobj = io.BytesIO()
    img2 = _mgh_rt(img, img_fobj)
    hdr2 = img2.header
    RZS = aff[:3, :3]
    zooms = np.sqrt(np.sum(RZS**2, axis=0))
    assert_almost_equal(hdr2['Mdc'].T, RZS / zooms)
    assert_almost_equal(hdr2['delta'], zooms)


def test_eq():
    # Test headers compare properly
    hdr = MGHHeader()
    hdr2 = MGHHeader()
    assert hdr == hdr2
    hdr.set_data_shape((2, 3, 4))
    assert hdr != hdr2
    hdr2.set_data_shape((2, 3, 4))
    assert hdr == hdr2


def test_header_slope_inter():
    # Test placeholder slope / inter method
    hdr = MGHHeader()
    assert hdr.get_slope_inter() == (None, None)


def test_mgh_load_fileobj():
    # Checks the filename gets passed to array proxy
    #
    # This is a bit of an implementation detail, but the test is to make sure
    # that we aren't passing ImageOpener objects to the array proxy, as these
    # were confusing mmap on Python 3.  If there's some sensible reason not to
    # pass the filename to the array proxy, please feel free to change this
    # test.
    img = MGHImage.load(MGZ_FNAME)
    assert pathlib.Path(img.dataobj.file_like) == pathlib.Path(MGZ_FNAME)
    # Check fileobj also passed into dataobj
    with ImageOpener(MGZ_FNAME) as fobj:
        contents = fobj.read()
    bio = io.BytesIO(contents)
    fm = MGHImage.make_file_map(mapping=dict(image=bio))
    img2 = MGHImage.from_file_map(fm)
    assert img2.dataobj.file_like is bio
    assert_array_equal(img.get_fdata(), img2.get_fdata())


def test_mgh_affine_default():
    hdr = MGHHeader()
    hdr['goodRASFlag'] = 0
    hdr2 = MGHHeader(hdr.binaryblock)
    assert hdr2['goodRASFlag'] == 1
    assert_array_equal(hdr['Mdc'], hdr2['Mdc'])
    assert_array_equal(hdr['Pxyz_c'], hdr2['Pxyz_c'])


def test_mgh_set_data_shape():
    hdr = MGHHeader()
    hdr.set_data_shape((5,))
    assert_array_equal(hdr.get_data_shape(), (5, 1, 1))
    hdr.set_data_shape((5, 4))
    assert_array_equal(hdr.get_data_shape(), (5, 4, 1))
    hdr.set_data_shape((5, 4, 3))
    assert_array_equal(hdr.get_data_shape(), (5, 4, 3))
    hdr.set_data_shape((5, 4, 3, 2))
    assert_array_equal(hdr.get_data_shape(), (5, 4, 3, 2))
    with pytest.raises(ValueError):
        hdr.set_data_shape((5, 4, 3, 2, 1))


def test_mghheader_default_structarr():
    hdr = MGHHeader.default_structarr()
    assert hdr['version'] == 1
    assert_array_equal(hdr['dims'], 1)
    assert hdr['type'] == 3
    assert hdr['dof'] == 0
    assert hdr['goodRASFlag'] == 1
    assert_array_equal(hdr['delta'], 1)
    assert_array_equal(hdr['Mdc'], [[-1, 0, 0], [0, 0, 1], [0, -1, 0]])
    assert_array_equal(hdr['Pxyz_c'], 0)
    assert hdr['tr'] == 0
    assert hdr['flip_angle'] == 0
    assert hdr['te'] == 0
    assert hdr['ti'] == 0
    assert hdr['fov'] == 0

    for endianness in (None,) + BIG_CODES:
        hdr2 = MGHHeader.default_structarr(endianness=endianness)
        assert hdr2 == hdr
        assert hdr2.view(hdr2.dtype.newbyteorder('>')) == hdr

    for endianness in LITTLE_CODES:
        with pytest.raises(ValueError):
            MGHHeader.default_structarr(endianness=endianness)


class TestMGHImage(tsi.TestSpatialImage, tsi.MmapImageMixin):
    """Apply general image tests to MGHImage"""

    image_class = MGHImage
    can_save = True

    def check_dtypes(self, expected, actual):
        # Some images will want dtypes to be equal including endianness,
        # others may only require the same type
        # MGH requires the actual to be a big endian version of expected
        assert expected.newbyteorder('>') == actual


class TestMGHHeader(tws._TestLabeledWrapStruct):
    header_class = MGHHeader

    def _set_something_into_hdr(self, hdr):
        hdr['dims'] = [4, 3, 2, 1]

    def get_bad_bb(self):
        return b'\xff' + b'\x00' * self.header_class._hdrdtype.itemsize

    # Update tests to account for big-endian requirement
    def test_general_init(self):
        hdr = self.header_class()
        # binaryblock has length given by header data dtype
        binblock = hdr.binaryblock
        assert len(binblock) == hdr.structarr.dtype.itemsize
        # Endianness will always be big, and cannot be set
        assert hdr.endianness == '>'
        # You can also pass in a check flag, without data this has no
        # effect
        hdr = self.header_class(check=False)

    def test__eq__(self):
        # Test equal and not equal
        hdr1 = self.header_class()
        hdr2 = self.header_class()
        assert hdr1 == hdr2
        self._set_something_into_hdr(hdr1)
        assert hdr1 != hdr2
        self._set_something_into_hdr(hdr2)
        assert hdr1 == hdr2
        # REMOVED as_byteswapped() test
        # Check comparing to funny thing says no
        assert hdr1 != None
        assert hdr1 != 1

    def test_to_from_fileobj(self):
        # Successful write using write_to
        hdr = self.header_class()
        str_io = io.BytesIO()
        hdr.write_to(str_io)
        str_io.seek(0)
        hdr2 = self.header_class.from_fileobj(str_io)
        assert hdr2.endianness == '>'
        assert hdr2.binaryblock == hdr.binaryblock

    def test_endian_guess(self):
        # Check guesses of endian
        eh = self.header_class()
        assert eh.endianness == '>'
        assert self.header_class.guessed_endian(eh) == '>'

    def test_bytes(self):
        # Test get of bytes
        hdr1 = self.header_class()
        bb = hdr1.binaryblock
        hdr2 = self.header_class(hdr1.binaryblock)
        assert hdr1 == hdr2
        assert hdr1.binaryblock == hdr2.binaryblock
        # Do a set into the header, and try again.  The specifics of 'setting
        # something' will depend on the nature of the bytes object
        self._set_something_into_hdr(hdr1)
        hdr2 = self.header_class(hdr1.binaryblock)
        assert hdr1 == hdr2
        assert hdr1.binaryblock == hdr2.binaryblock
        # Short binaryblocks give errors (here set through init)
        # Long binaryblocks are truncated
        with pytest.raises(WrapStructError):
            self.header_class(bb[: self.header_class._hdrdtype.itemsize - 1])

        # Checking set to true by default, and prevents nonsense being
        # set into the header.
        bb_bad = self.get_bad_bb()
        if bb_bad is None:
            return
        with imageglobals.LoggingOutputSuppressor():
            with pytest.raises(HeaderDataError):
                self.header_class(bb_bad)

        # now slips past without check
        _ = self.header_class(bb_bad, check=False)

    def test_as_byteswapped(self):
        # Check byte swapping
        hdr = self.header_class()
        assert hdr.endianness == '>'
        # same code just returns a copy
        for endianness in BIG_CODES:
            hdr2 = hdr.as_byteswapped(endianness)
            assert hdr2 is not hdr
            assert hdr2 == hdr

        # Different code raises error
        for endianness in (None,) + LITTLE_CODES:
            with pytest.raises(ValueError):
                hdr.as_byteswapped(endianness)

        # Note that contents is not rechecked on swap / copy
        class DC(self.header_class):
            def check_fix(self, *args, **kwargs):
                raise Exception

        # Assumes check=True default
        with pytest.raises(Exception):
            DC(hdr.binaryblock)

        hdr = DC(hdr.binaryblock, check=False)
        hdr2 = hdr.as_byteswapped('>')

    def test_checks(self):
        # Test header checks
        hdr_t = self.header_class()
        # _dxer just returns the diagnostics as a string
        # Default hdr is OK
        assert self._dxer(hdr_t) == ''
        # Version should be 1
        hdr = hdr_t.copy()
        hdr['version'] = 2
        assert self._dxer(hdr) == 'Unknown MGH format version'



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for Euler angles"""

import math

import numpy as np
import pytest
from numpy import pi
from numpy.testing import assert_array_almost_equal, assert_array_equal

from .. import eulerangles as nea
from .. import quaternions as nq

FLOAT_EPS = np.finfo(np.float64).eps

# Example rotations """
params = np.arange(-pi * 2, pi * 2.5, pi / 2)
eg_rots = [(x, y, z) for x in params for y in params for z in params]


def x_only(x):
    cosx = np.cos(x)
    sinx = np.sin(x)
    return np.array(
        [
            [1, 0, 0],
            [0, cosx, -sinx],
            [0, sinx, cosx],
        ]
    )


def y_only(y):
    cosy = np.cos(y)
    siny = np.sin(y)
    return np.array(
        [
            [cosy, 0, siny],
            [0, 1, 0],
            [-siny, 0, cosy],
        ]
    )


def z_only(z):
    cosz = np.cos(z)
    sinz = np.sin(z)
    return np.array(
        [
            [cosz, -sinz, 0],
            [sinz, cosz, 0],
            [0, 0, 1],
        ]
    )


def sympy_euler(z, y, x):
    # The whole matrix formula for z,y,x rotations from Sympy
    cos = math.cos
    sin = math.sin
    # the following copy / pasted from Sympy - see derivations subdirectory
    return [
        [cos(y) * cos(z), -cos(y) * sin(z), sin(y)],
        [
            cos(x) * sin(z) + cos(z) * sin(x) * sin(y),
            cos(x) * cos(z) - sin(x) * sin(y) * sin(z),
            -cos(y) * sin(x),
        ],
        [
            sin(x) * sin(z) - cos(x) * cos(z) * sin(y),
            cos(z) * sin(x) + cos(x) * sin(y) * sin(z),
            cos(x) * cos(y),
        ],
    ]


def is_valid_rotation(M):
    if not np.allclose(np.linalg.det(M), 1):
        return False
    return np.allclose(np.eye(3), np.dot(M, M.T))


def test_basic_euler():
    # some example rotations, in radians
    zr = 0.05
    yr = -0.4
    xr = 0.2
    # Rotation matrix composing the three rotations
    M = nea.euler2mat(zr, yr, xr)
    # Corresponding individual rotation matrices
    M1 = nea.euler2mat(zr)
    M2 = nea.euler2mat(0, yr)
    M3 = nea.euler2mat(0, 0, xr)
    # which are all valid rotation matrices
    assert is_valid_rotation(M)
    assert is_valid_rotation(M1)
    assert is_valid_rotation(M2)
    assert is_valid_rotation(M3)
    # Full matrix is composition of three individual matrices
    assert np.allclose(M, np.dot(M3, np.dot(M2, M1)))
    # Rotations can be specified with named args, default 0
    assert np.all(nea.euler2mat(zr) == nea.euler2mat(z=zr))
    assert np.all(nea.euler2mat(0, yr) == nea.euler2mat(y=yr))
    assert np.all(nea.euler2mat(0, 0, xr) == nea.euler2mat(x=xr))
    # Applying an opposite rotation same as inverse (the inverse is
    # the same as the transpose, but just for clarity)
    assert np.allclose(nea.euler2mat(x=-xr), np.linalg.inv(nea.euler2mat(x=xr)))


def test_euler_mat_1():
    M = nea.euler2mat()
    assert_array_equal(M, np.eye(3))


@pytest.mark.parametrize(('x', 'y', 'z'), eg_rots)
def test_euler_mat_2(x, y, z):
    M1 = nea.euler2mat(z, y, x)
    M2 = sympy_euler(z, y, x)
    assert_array_almost_equal(M1, M2)
    M3 = np.dot(x_only(x), np.dot(y_only(y), z_only(z)))
    assert_array_almost_equal(M1, M3)
    zp, yp, xp = nea.mat2euler(M1)
    # The parameters may not be the same as input, but they give the
    # same rotation matrix
    M4 = nea.euler2mat(zp, yp, xp)
    assert_array_almost_equal(M1, M4)


def sympy_euler2quat(z=0, y=0, x=0):
    # direct formula for z,y,x quaternion rotations using sympy
    # see derivations subfolder
    cos = math.cos
    sin = math.sin
    # the following copy / pasted from Sympy output
    return (
        cos(0.5 * x) * cos(0.5 * y) * cos(0.5 * z) - sin(0.5 * x) * sin(0.5 * y) * sin(0.5 * z),
        cos(0.5 * x) * sin(0.5 * y) * sin(0.5 * z) + cos(0.5 * y) * cos(0.5 * z) * sin(0.5 * x),
        cos(0.5 * x) * cos(0.5 * z) * sin(0.5 * y) - cos(0.5 * y) * sin(0.5 * x) * sin(0.5 * z),
        cos(0.5 * x) * cos(0.5 * y) * sin(0.5 * z) + cos(0.5 * z) * sin(0.5 * x) * sin(0.5 * y),
    )


def crude_mat2euler(M):
    """The simplest possible - ignoring atan2 instability"""
    r11, r12, r13, r21, r22, r23, r31, r32, r33 = M.flat
    return math.atan2(-r12, r11), math.asin(r13), math.atan2(-r23, r33)


def test_euler_instability():
    # Test for numerical errors in mat2euler
    # problems arise for cos(y) near 0
    po2 = pi / 2
    zyx = po2, po2, po2
    M = nea.euler2mat(*zyx)
    # Round trip
    M_back = nea.euler2mat(*nea.mat2euler(M))
    assert np.allclose(M, M_back)
    # disturb matrix slightly
    M_e = M - FLOAT_EPS
    # round trip to test - OK
    M_e_back = nea.euler2mat(*nea.mat2euler(M_e))
    assert np.allclose(M_e, M_e_back)
    # not so with crude routine
    M_e_back = nea.euler2mat(*crude_mat2euler(M_e))
    assert not np.allclose(M_e, M_e_back)


@pytest.mark.parametrize(('x', 'y', 'z'), eg_rots)
def test_quats(x, y, z):
    M1 = nea.euler2mat(z, y, x)
    quatM = nq.mat2quat(M1)
    quat = nea.euler2quat(z, y, x)
    assert nq.nearly_equivalent(quatM, quat)
    quatS = sympy_euler2quat(z, y, x)
    assert nq.nearly_equivalent(quat, quatS)
    zp, yp, xp = nea.quat2euler(quat)
    # The parameters may not be the same as input, but they give the
    # same rotation matrix
    M2 = nea.euler2mat(zp, yp, xp)
    assert_array_almost_equal(M1, M2)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Test scripts

Test running scripts
"""

from io import StringIO
from os.path import join as pjoin

import numpy as np
import pytest

import nibabel as nib
from nibabel.cmdline.diff import (
    display_diff,
    get_data_diff,
    get_data_hash_diff,
    get_headers_diff,
    main,
)
from nibabel.cmdline.utils import (
    ap,
    safe_get,
    table2string,
)
from nibabel.testing import data_path


def test_table2string():
    # Trivial case should do something sensible
    assert table2string([]) == '\n'
    assert (
        table2string(
            [['A', 'B', 'C', 'D'],
             ['E', 'F', 'G', 'H']]
        ) == (
            'A B C D\n'
            'E F G H\n'
        )
    )  # fmt: skip
    assert (
        table2string(
            [["Let's", 'Make', 'Tests', 'And'],
             ['Have', 'Lots', 'Of', 'Fun'],
             ['With', 'Python', 'Guys', '!']]
        ) == (
            "Let's  Make  Tests And\n"
            'Have   Lots   Of   Fun\n'
            'With  Python Guys   !\n'
        )
    )  # fmt: skip
    assert (
        table2string(
            [['This', 'Table', '@lIs', 'Ragged'],
             ['And', '@rit', 'uses', '@csome', 'alignment', 'markup']]
        ) == (
            'This Table Is   Ragged\n'
            'And     it uses  some  alignment markup\n'
        )
    )  # fmt: skip


def test_ap():
    assert ap([1, 2], '%2d') == ' 1,  2'
    assert ap([1, 2], '%3d') == '  1,   2'
    assert ap([1, 2], '%-2d') == '1 , 2 '
    assert ap([1, 2], '%d', '+') == '1+2'
    assert ap([1, 2, 3], '%d', '-') == '1-2-3'


def test_safe_get():
    class TestObject:
        def __init__(self, test=None):
            self.test = test

        def get_test(self):
            return self.test

    test = TestObject()
    test.test = 2

    assert safe_get(test, 'test') == 2
    assert safe_get(test, 'failtest') == '-'


def test_get_headers_diff():
    fnames = [pjoin(data_path, f) for f in ('standard.nii.gz', 'example4d.nii.gz')]
    actual_difference = get_headers_diff([nib.load(f).header for f in fnames])
    expected_difference = {
        'regular': [np.asarray(b''), np.asarray(b'r')],
        'dim_info': [np.asarray(0, 'uint8'), np.asarray(57, 'uint8')],
        'dim': [
            np.array([3, 4, 5, 7, 1, 1, 1, 1], 'int16'),
            np.array([4, 128, 96, 24, 2, 1, 1, 1], 'int16'),
        ],
        'datatype': [np.array(2, 'uint8'), np.array(4, 'uint8')],
        'bitpix': [np.array(8, 'uint8'), np.array(16, 'uint8')],
        'pixdim': [
            np.array([1.0, 1.0, 3.0, 2.0, 1.0, 1.0, 1.0, 1.0], 'float32'),
            np.array(
                [
                    -1.00000000e00,
                    2.00000000e00,
                    2.00000000e00,
                    2.19999909e00,
                    2.00000000e03,
                    1.00000000e00,
                    1.00000000e00,
                    1.00000000e00,
                ],
                'float32',
            ),
        ],
        'slice_end': [np.array(0, 'uint8'), np.array(23, 'uint8')],
        'xyzt_units': [np.array(0, 'uint8'), np.array(10, 'uint8')],
        'cal_max': [
            np.array(0.0, 'float32'),
            np.asarray(1162.0, 'float32'),
        ],
        'descrip': [
            np.array(b'', 'S80'),
            np.array(b'FSL3.3\x00 v2.25 NIfTI-1 Single file format', 'S80'),
        ],
        'qform_code': [np.array(0, 'int16'), np.array(1, 'int16')],
        'sform_code': [np.array(2, 'int16'), np.array(1, 'int16')],
        'quatern_b': [
            np.array(0.0, 'float32'),
            np.array(-1.9451068140294884e-26, 'float32'),
        ],
        'quatern_c': [
            np.array(0.0, 'float32'),
            np.array(-0.9967085123062134, 'float32'),
        ],
        'quatern_d': [
            np.array(0.0, 'float32'),
            np.array(-0.0810687392950058, 'float32'),
        ],
        'qoffset_x': [
            np.array(0.0, 'float32'),
            np.array(117.8551025390625, 'float32'),
        ],
        'qoffset_y': [
            np.array(0.0, 'float32'),
            np.array(-35.72294235229492, 'float32'),
        ],
        'qoffset_z': [
            np.array(0.0, 'float32'),
            np.array(-7.248798370361328, 'float32'),
        ],
        'srow_x': [
            np.array([1.0, 0.0, 0.0, 0.0], 'float32'),
            np.array([-2.00000000e00, 6.71471565e-19, 9.08102451e-18, 1.17855103e02], 'float32'),
        ],
        'srow_y': [
            np.array([0.0, 3.0, 0.0, 0.0], 'float32'),
            np.array([-6.71471565e-19, 1.97371149e00, -3.55528235e-01, -3.57229424e01], 'float32'),
        ],
        'srow_z': [
            np.array([0.0, 0.0, 2.0, 0.0], 'float32'),
            np.array([8.25548089e-18, 3.23207617e-01, 2.17108178e00, -7.24879837e00], 'float32'),
        ],
    }

    np.testing.assert_equal(actual_difference, expected_difference)


def test_display_diff():
    bogus_names = ['hellokitty.nii.gz', 'privettovarish.nii.gz']

    dict_values = {
        'datatype': [np.array(2, 'uint8'), np.array(4, 'uint8')],
        'bitpix': [np.array(8, 'uint8'), np.array(16, 'uint8')],
    }

    expected_output = """\
These files are different.
Field/File     \
1:hellokitty.nii.gz                                    \
2:privettovarish.nii.gz                                \n\
datatype       \
2                                                      \
4                                                      \n\
bitpix         \
8                                                      \
16                                                     \n"""

    assert display_diff(bogus_names, dict_values) == expected_output


def test_get_data_diff():
    #  testing for identical files specifically as md5 may vary by computer
    test_names = [pjoin(data_path, f) for f in ('standard.nii.gz', 'standard.nii.gz')]
    assert get_data_hash_diff(test_names) == []

    #  testing the maximum relative and absolute differences' different use cases
    test_array = np.arange(16).reshape(4, 4)
    test_array_2 = np.arange(1, 17).reshape(4, 4)
    test_array_3 = np.arange(2, 18).reshape(4, 4)
    test_array_4 = np.arange(100).reshape(10, 10)
    test_array_5 = np.arange(64).reshape(8, 8)

    # same shape, 2 files
    assert get_data_diff([test_array, test_array_2]) == {
        'DATA(diff 1:)': [None, {'abs': 1, 'rel': 2.0}]
    }

    # same shape, 3 files
    assert get_data_diff([test_array, test_array_2, test_array_3]) == {
        'DATA(diff 1:)': [
            None,
            {'abs': 1, 'rel': 2.0},
            {'abs': 2, 'rel': 2.0},
        ],
        'DATA(diff 2:)': [None, None, {'abs': 1, 'rel': 0.66666666666666663}],
    }

    # same shape, 2 files, modified maximum abs/rel
    assert get_data_diff([test_array, test_array_2], max_abs=2, max_rel=2) == {}

    # different shape, 2 files
    assert get_data_diff([test_array_2, test_array_4]) == {
        'DATA(diff 1:)': [None, {'CMP': 'incompat'}]
    }

    # different shape, 3 files
    assert get_data_diff([test_array_4, test_array_5, test_array_2]) == {
        'DATA(diff 1:)': [None, {'CMP': 'incompat'}, {'CMP': 'incompat'}],
        'DATA(diff 2:)': [None, None, {'CMP': 'incompat'}],
    }

    test_return = get_data_diff([test_array, test_array_2], dtype=np.float32)
    assert type(test_return['DATA(diff 1:)'][1]['abs']) is np.float32
    assert type(test_return['DATA(diff 1:)'][1]['rel']) is np.float32

    test_return_2 = get_data_diff([test_array, test_array_2, test_array_3])
    assert type(test_return_2['DATA(diff 1:)'][1]['abs']) is np.float64
    assert type(test_return_2['DATA(diff 1:)'][1]['rel']) is np.float64
    assert type(test_return_2['DATA(diff 2:)'][2]['abs']) is np.float64
    assert type(test_return_2['DATA(diff 2:)'][2]['rel']) is np.float64


def test_main():
    test_names = [pjoin(data_path, f) for f in ('standard.nii.gz', 'example4d.nii.gz')]
    expected_difference = {
        'regular': [np.asarray(b''), np.asarray(b'r')],
        'dim_info': [np.asarray(0, 'uint8'), np.asarray(57, 'uint8')],
        'dim': [
            np.array([3, 4, 5, 7, 1, 1, 1, 1], 'int16'),
            np.array([4, 128, 96, 24, 2, 1, 1, 1], 'int16'),
        ],
        'datatype': [np.array(2, 'uint8'), np.array(4, 'uint8')],
        'bitpix': [np.array(8, 'uint8'), np.array(16, 'uint8')],
        'pixdim': [
            np.array([1.0, 1.0, 3.0, 2.0, 1.0, 1.0, 1.0, 1.0], 'float32'),
            np.array(
                [
                    -1.00000000e00,
                    2.00000000e00,
                    2.00000000e00,
                    2.19999909e00,
                    2.00000000e03,
                    1.00000000e00,
                    1.00000000e00,
                    1.00000000e00,
                ],
                'float32',
            ),
        ],
        'slice_end': [np.array(0, 'uint8'), np.array(23, 'uint8')],
        'xyzt_units': [np.array(0, 'uint8'), np.array(10, 'uint8')],
        'cal_max': [
            np.array(0.0, 'float32'),
            np.asarray(1162.0, 'float32'),
        ],
        'descrip': [
            np.array(b'', 'S80'),
            np.array(b'FSL3.3\x00 v2.25 NIfTI-1 Single file format', 'S80'),
        ],
        'qform_code': [np.array(0, 'int16'), np.array(1, 'int16')],
        'sform_code': [np.array(2, 'int16'), np.array(1, 'int16')],
        'quatern_b': [
            np.array(0.0, 'float32'),
            np.array(-1.9451068140294884e-26, 'float32'),
        ],
        'quatern_c': [
            np.array(0.0, 'float32'),
            np.array(-0.9967085123062134, 'float32'),
        ],
        'quatern_d': [
            np.array(0.0, 'float32'),
            np.array(-0.0810687392950058, 'float32'),
        ],
        'qoffset_x': [
            np.array(0.0, 'float32'),
            np.array(117.8551025390625, 'float32'),
        ],
        'qoffset_y': [
            np.array(0.0, 'float32'),
            np.array(-35.72294235229492, 'float32'),
        ],
        'qoffset_z': [
            np.array(0.0, 'float32'),
            np.array(-7.248798370361328, 'float32'),
        ],
        'srow_x': [
            np.array([1.0, 0.0, 0.0, 0.0], 'float32'),
            np.array([-2.00000000e00, 6.71471565e-19, 9.08102451e-18, 1.17855103e02], 'float32'),
        ],
        'srow_y': [
            np.array([0.0, 3.0, 0.0, 0.0], 'float32'),
            np.array([-6.71471565e-19, 1.97371149e00, -3.55528235e-01, -3.57229424e01], 'float32'),
        ],
        'srow_z': [
            np.array([0.0, 0.0, 2.0, 0.0], 'float32'),
            np.array([8.25548089e-18, 3.23207617e-01, 2.17108178e00, -7.24879837e00], 'float32'),
        ],
        'DATA(md5)': ['0a2576dd6badbb25bfb3b12076df986b', 'b0abbc492b4fd533b2c80d82570062cf'],
    }

    with pytest.raises(SystemExit):
        np.testing.assert_equal(main(test_names, StringIO()), expected_difference)

    test_names_2 = [pjoin(data_path, f) for f in ('standard.nii.gz', 'standard.nii.gz')]

    with pytest.raises(SystemExit):
        assert main(test_names_2, StringIO()) == 'These files are identical.'



"""Testing filebasedimages module"""

import warnings
from itertools import product

import numpy as np
import pytest

from ..filebasedimages import FileBasedHeader, FileBasedImage, SerializableImage
from .test_image_api import GenericImageAPI, SerializeMixin


class FBNumpyImage(FileBasedImage):
    header_class = FileBasedHeader
    valid_exts = ('.npy',)
    files_types = (('image', '.npy'),)

    def __init__(self, arr, header=None, extra=None, file_map=None):
        super().__init__(header, extra, file_map)
        self.arr = arr

    @property
    def shape(self):
        return self.arr.shape

    def get_data(self):
        warnings.warn('Deprecated', DeprecationWarning)
        return self.arr

    @property
    def dataobj(self):
        return self.arr

    def get_fdata(self):
        return self.arr.astype(np.float64)

    @classmethod
    def from_file_map(klass, file_map):
        with file_map['image'].get_prepare_fileobj('rb') as fobj:
            arr = np.load(fobj)
        return klass(arr)

    def to_file_map(self, file_map=None):
        file_map = self.file_map if file_map is None else file_map
        with file_map['image'].get_prepare_fileobj('wb') as fobj:
            np.save(fobj, self.arr)

    def get_data_dtype(self):
        return self.arr.dtype

    def set_data_dtype(self, dtype):
        self.arr = self.arr.astype(dtype)


class SerializableNumpyImage(FBNumpyImage, SerializableImage):
    pass


class TestFBImageAPI(GenericImageAPI):
    """Validation for FileBasedImage instances"""

    # A callable returning an image from ``image_maker(data, header)``
    image_maker = FBNumpyImage
    # A callable returning a header from ``header_maker()``
    header_maker = FileBasedHeader
    # Example shapes for created images
    example_shapes = ((2,), (2, 3), (2, 3, 4), (2, 3, 4, 5))
    example_dtypes = (np.int8, np.uint16, np.int32, np.float32)
    can_save = True
    standard_extension = '.npy'

    def make_imaker(self, arr, header=None):
        return lambda: self.image_maker(arr, header)

    def obj_params(self):
        # Create new images
        for shape, dtype in product(self.example_shapes, self.example_dtypes):
            arr = np.arange(np.prod(shape), dtype=dtype).reshape(shape)
            hdr = self.header_maker()
            func = self.make_imaker(arr.copy(), hdr)
            params = dict(dtype=dtype, data=arr, shape=shape, is_proxy=False)
            yield func, params


class TestSerializableImageAPI(TestFBImageAPI, SerializeMixin):
    image_maker = SerializableNumpyImage

    @staticmethod
    def _header_eq(header_a, header_b):
        """FileBasedHeader is an abstract class, so __eq__ is undefined.
        Checking for the same header type is sufficient, here."""
        return type(header_a) == type(header_b) == FileBasedHeader


def test_filebased_header():
    # Test stuff about the default FileBasedHeader

    class H(FileBasedHeader):
        def __init__(self, seq=None):
            if seq is None:
                seq = []
            self.a_list = list(seq)

    in_list = [1, 3, 2]
    hdr = H(in_list)
    hdr_c = hdr.copy()
    assert hdr_c.a_list == hdr.a_list
    # Copy is independent of original
    hdr_c.a_list[0] = 99
    assert hdr_c.a_list != hdr.a_list
    # From header does a copy
    hdr2 = H.from_header(hdr)
    assert isinstance(hdr2, H)
    assert hdr2.a_list == hdr.a_list
    hdr2.a_list[0] = 42
    assert hdr2.a_list != hdr.a_list
    # Default header input to from_heder gives new empty header
    hdr3 = H.from_header()
    assert isinstance(hdr3, H)
    assert hdr3.a_list == []
    hdr4 = H.from_header(None)
    assert isinstance(hdr4, H)
    assert hdr4.a_list == []


class MultipartNumpyImage(FBNumpyImage):
    # We won't actually try to write these out, just need to test an edge case
    files_types = (('header', '.hdr'), ('image', '.npy'))


class SerializableMPNumpyImage(MultipartNumpyImage, SerializableImage):
    pass


def test_multifile_stream_failure():
    shape = (2, 3, 4)
    arr = np.arange(np.prod(shape), dtype=np.float32).reshape(shape)
    img = SerializableMPNumpyImage(arr)
    with pytest.raises(NotImplementedError):
        img.to_bytes()
    img = SerializableNumpyImage(arr)
    bstr = img.to_bytes()
    with pytest.raises(NotImplementedError):
        SerializableMPNumpyImage.from_bytes(bstr)



"""Module to help with deprecating objects and classes"""

from __future__ import annotations

import typing as ty
import warnings

from ._typing import ParamSpec
from .deprecator import Deprecator
from .pkg_info import cmp_pkg_version

P = ParamSpec('P')


class ModuleProxy:
    """Proxy for module that may not yet have been imported

    Parameters
    ----------
    module_name : str
        Full module name e.g. ``nibabel.minc``

    Examples
    --------

    ::
        arr = np.arange(24).reshape((2, 3, 4))
        nifti1 = ModuleProxy('nibabel.nifti1')
        nifti1_image = nifti1.Nifti1Image(arr, np.eye(4))

    So, the ``nifti1`` object is a proxy that will import the required module
    when you do attribute access and return the attributes of the imported
    module.
    """

    def __init__(self, module_name: str) -> None:
        self._module_name = module_name

    def __getattr__(self, key: str) -> ty.Any:
        mod = __import__(self._module_name, fromlist=[''])
        return getattr(mod, key)

    def __repr__(self) -> str:
        return f'<module proxy for {self._module_name}>'


class FutureWarningMixin(ty.Generic[P]):
    """Insert FutureWarning for object creation

    Examples
    --------
    >>> class C: pass
    >>> class D(FutureWarningMixin, C):
    ...     warn_message = "Please, don't use this class"

    Record the warning

    >>> with warnings.catch_warnings(record=True) as warns:
    ...     d = D()
    ...     warns[0].message.args[0]
    "Please, don't use this class"
    """

    warn_message = 'This class will be removed in future versions'

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        warnings.warn(self.warn_message, FutureWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class VisibleDeprecationWarning(UserWarning):
    """Deprecation warning that will be shown by default

    Python >= 2.7 does not show standard DeprecationWarnings by default:

    http://docs.python.org/dev/whatsnew/2.7.html#the-future-for-python-2-x

    Use this class for cases where we do want to show deprecations by default.
    """

    pass


deprecate_with_version = Deprecator(cmp_pkg_version)


def alert_future_error(
    msg: str,
    version: str,
    *,
    warning_class: type[Warning] = FutureWarning,
    error_class: type[Exception] = RuntimeError,
    warning_rec: str = '',
    error_rec: str = '',
    stacklevel: int = 2,
) -> None:
    """Warn or error with appropriate messages for changing functionality.

    Parameters
    ----------
    msg : str
        Description of the condition that led to the alert
    version : str
        NiBabel version at which the warning will become an error
    warning_class : subclass of Warning, optional
        Warning class to emit before version
    error_class : subclass of Exception, optional
        Error class to emit after version
    warning_rec : str, optional
        Guidance for suppressing the warning and avoiding the future error
    error_rec: str, optional
        Guidance for resolving the error
    stacklevel: int, optional
        Warnings stacklevel to provide; note that this will be incremented by
        1, so provide the stacklevel you would provide directly to warnings.warn()
    """
    if cmp_pkg_version(version) > 0:
        msg = f'{msg} This will error in NiBabel {version}. {warning_rec}'
        warnings.warn(msg.strip(), warning_class, stacklevel=stacklevel + 1)
    else:
        raise error_class(f'{msg} {error_rec}'.strip())



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test for scaling / rounding in volumeutils module"""

import warnings
from io import BytesIO

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from ..casting import sctypes, type_info
from ..testing import suppress_warnings
from ..volumeutils import apply_read_scaling, array_from_file, array_to_file, finite_range
from .test_volumeutils import _calculate_scale

# Debug print statements
DEBUG = True


@pytest.mark.parametrize(
    ('in_arr', 'res'),
    [
        ([[-1, 0, 1], [np.inf, np.nan, -np.inf]], (-1, 1)),
        (np.array([[-1, 0, 1], [np.inf, np.nan, -np.inf]]), (-1, 1)),
        ([[np.nan], [np.nan]], (np.inf, -np.inf)),  # all nans slices
        (np.zeros((3, 4, 5)) + np.nan, (np.inf, -np.inf)),
        ([[-np.inf], [np.inf]], (np.inf, -np.inf)),  # all infs slices
        (np.zeros((3, 4, 5)) + np.inf, (np.inf, -np.inf)),
        ([[np.nan, -1, 2], [-2, np.nan, 1]], (-2, 2)),
        ([[np.nan, -np.inf, 2], [-2, np.nan, np.inf]], (-2, 2)),
        ([[-np.inf, 2], [np.nan, 1]], (1, 2)),  # good max case
        ([np.nan], (np.inf, -np.inf)),
        ([np.inf], (np.inf, -np.inf)),
        ([-np.inf], (np.inf, -np.inf)),
        ([np.inf, 1], (1, 1)),  # only look at finite values
        ([-np.inf, 1], (1, 1)),
        ([[], []], (np.inf, -np.inf)),  # empty array
        (np.array([[-3, 0, 1], [2, -1, 4]], dtype=int), (-3, 4)),
        (np.array([[1, 0, 1], [2, 3, 4]], dtype=np.uint), (0, 4)),
        ([0.0, 1, 2, 3], (0, 3)),
        # Complex comparison works as if they are floats
        ([[np.nan, -1 - 100j, 2], [-2, np.nan, 1 + 100j]], (-2, 2)),
        ([[np.nan, -1, 2 - 100j], [-2 + 100j, np.nan, 1]], (-2 + 100j, 2 - 100j)),
    ],
)
def test_finite_range(in_arr, res):
    # Finite range utility function
    assert finite_range(in_arr) == res
    assert finite_range(in_arr, False) == res
    assert finite_range(in_arr, check_nan=False) == res
    has_nan = np.any(np.isnan(in_arr))
    assert finite_range(in_arr, True) == res + (has_nan,)
    assert finite_range(in_arr, check_nan=True) == res + (has_nan,)
    in_arr = np.array(in_arr)
    flat_arr = in_arr.ravel()
    assert finite_range(flat_arr) == res
    assert finite_range(flat_arr, True) == res + (has_nan,)
    # Check float types work as complex
    if in_arr.dtype.kind == 'f':
        c_arr = in_arr.astype(np.complex128)
        assert finite_range(c_arr) == res
        assert finite_range(c_arr, True) == res + (has_nan,)


def test_finite_range_err():
    # Test error cases
    a = np.array([[1.0, 0, 1], [2, 3, 4]]).view([('f1', 'f')])
    with pytest.raises(TypeError):
        finite_range(a)


@pytest.mark.parametrize('out_type', [np.int16, np.float32])
def test_a2f_mn_mx(out_type):
    # Test array to file mn, mx handling
    str_io = BytesIO()
    arr = np.arange(6, dtype=out_type)
    arr_orig = arr.copy()  # safe backup for testing against
    # Basic round trip to warm up
    array_to_file(arr, str_io)
    data_back = array_from_file(arr.shape, out_type, str_io)
    assert_array_equal(arr, data_back)
    # Clip low
    array_to_file(arr, str_io, mn=2)
    data_back = array_from_file(arr.shape, out_type, str_io)
    # arr unchanged
    assert_array_equal(arr, arr_orig)
    # returned value clipped low
    assert_array_equal(data_back, [2, 2, 2, 3, 4, 5])
    # Clip high
    array_to_file(arr, str_io, mx=4)
    data_back = array_from_file(arr.shape, out_type, str_io)
    # arr unchanged
    assert_array_equal(arr, arr_orig)
    # returned value clipped high
    assert_array_equal(data_back, [0, 1, 2, 3, 4, 4])
    # Clip both
    array_to_file(arr, str_io, mn=2, mx=4)
    data_back = array_from_file(arr.shape, out_type, str_io)
    # arr unchanged
    assert_array_equal(arr, arr_orig)
    # returned value clipped high
    assert_array_equal(data_back, [2, 2, 2, 3, 4, 4])


def test_a2f_nan2zero():
    # Test conditions under which nans written to zero
    arr = np.array([np.nan, 99.0], dtype=np.float32)
    str_io = BytesIO()
    array_to_file(arr, str_io)
    data_back = array_from_file(arr.shape, np.float32, str_io)
    assert_array_equal(np.isnan(data_back), [True, False])
    # nan2zero ignored for floats
    array_to_file(arr, str_io, nan2zero=True)
    data_back = array_from_file(arr.shape, np.float32, str_io)
    assert_array_equal(np.isnan(data_back), [True, False])
    # Integer output with nan2zero gives zero
    with np.errstate(invalid='ignore'):
        array_to_file(arr, str_io, np.int32, nan2zero=True)
    data_back = array_from_file(arr.shape, np.int32, str_io)
    assert_array_equal(data_back, [0, 99])
    # Integer output with nan2zero=False gives whatever astype gives
    with np.errstate(invalid='ignore'):
        array_to_file(arr, str_io, np.int32, nan2zero=False)
    data_back = array_from_file(arr.shape, np.int32, str_io)
    assert_array_equal(data_back, [np.array(np.nan).astype(np.int32), 99])


@pytest.mark.parametrize(
    ('in_type', 'out_type'),
    [
        (np.int16, np.int16),
        (np.int16, np.int8),
        (np.uint16, np.uint8),
        (np.int32, np.int8),
        (np.float32, np.uint8),
        (np.float32, np.int16),
    ],
)
def test_array_file_scales(in_type, out_type):
    # Test scaling works for max, min when going from larger to smaller type,
    # and from float to integer.
    bio = BytesIO()
    out_dtype = np.dtype(out_type)
    arr = np.zeros((3,), dtype=in_type)
    info = type_info(in_type)
    arr[0], arr[1] = info['min'], info['max']
    slope, inter, mn, mx = _calculate_scale(arr, out_dtype, True)
    array_to_file(arr, bio, out_type, 0, inter, slope, mn, mx)
    bio.seek(0)
    arr2 = array_from_file(arr.shape, out_dtype, bio)
    arr3 = apply_read_scaling(arr2, slope, inter)
    # Max rounding error for integer type
    max_miss = slope / 2.0
    assert np.all(np.abs(arr - arr3) <= max_miss)


@pytest.mark.parametrize(
    ('category0', 'category1', 'overflow'),
    [
        # Confirm that, for all ints and uints as input, and all possible outputs,
        # for any simple way of doing the calculation, the result is near enough
        ('int', 'int', False),
        ('uint', 'int', False),
        # Converting floats to integer
        ('float', 'int', True),
        ('float', 'uint', True),
        ('complex', 'int', True),
        ('complex', 'uint', True),
    ],
)
def test_scaling_in_abstract(category0, category1, overflow):
    for in_type in sctypes[category0]:
        for out_type in sctypes[category1]:
            if overflow:
                with suppress_warnings():
                    check_int_a2f(in_type, out_type)
            else:
                check_int_a2f(in_type, out_type)


def check_int_a2f(in_type, out_type):
    # Check that array to / from file returns roughly the same as input
    big_floater = sctypes['float'][-1]
    info = type_info(in_type)
    this_min, this_max = info['min'], info['max']
    if not in_type in sctypes['complex']:
        data = np.array([this_min, this_max], in_type)
        # Bug in numpy 1.6.2 on PPC leading to infs - abort
        if not np.all(np.isfinite(data)):
            if DEBUG:
                print(f'Hit PPC max -> inf bug; skip in_type {in_type}')
            return
    else:  # Funny behavior with complex256
        data = np.zeros((2,), in_type)
        data[0] = this_min + 0j
        data[1] = this_max + 0j
    str_io = BytesIO()
    try:
        scale, inter, mn, mx = _calculate_scale(data, out_type, True)
    except ValueError as e:
        if DEBUG:
            warnings.warn(str((in_type, out_type, e)))
        return
    array_to_file(data, str_io, out_type, 0, inter, scale, mn, mx)
    data_back = array_from_file(data.shape, out_type, str_io)
    data_back = apply_read_scaling(data_back, scale, inter)
    assert np.allclose(big_floater(data), big_floater(data_back))
    # Try with analyze-size scale and inter
    scale32 = np.float32(scale)
    inter32 = np.float32(inter)
    if scale32 == np.inf or inter32 == np.inf:
        return
    data_back = array_from_file(data.shape, out_type, str_io)
    data_back = apply_read_scaling(data_back, scale32, inter32)
    # Clip at extremes to remove inf
    info = type_info(in_type)
    out_min, out_max = info['min'], info['max']
    assert np.allclose(big_floater(data), big_floater(np.clip(data_back, out_min, out_max)))
